// FitDip() -- fit one magnetic dipole to measurements from an array of OPM sensors
//
//  This subroutine employs the well-known Marquardt-Levenberg solution.
//  The position & orientation of a single magnetic dipole are fitted to measurements
//  from an array of OPM magnetometer sensors.
//
//  Author: Stephen E. Robinson
//          MEG Core Facility
//          NIMH
//

#include "jig.h"


double  FitDip(
    gsl_vector  *Xopm,          // calibrator measurements for array of OPM sensors & single dipole - N_OPM x 1 (input)
    FIELD       *Sensor,        // Sensor[N_OPM] - array of OPM sensors (input)
    double      *noise,         // noise[N_OPM] - measured noise (input)
    FIELD       *Dipole         // single magnetic dipole (initial guess & output)
)
{
    gsl_matrix      *covar;     // covar - 6x6 augmented covariance of partial derivatives
    gsl_matrix      *cov;       // cov - 6x6 covariance (copy)
    gsl_matrix      *alpha;     // alpha - 6x6 partial derivatives
    gsl_vector      *beta;      // beta - 6x1 vector
    gsl_vector      *da;        // input vector - 6x1
    gsl_vector      *x;         // solution vector - 6x1
    gsl_vector      *a;         // linear OPM parameters - 6x1
    gsl_vector      *aTry;      // trial parameters - 6x1
	gsl_permutation	*indx;      // LU decomposition permutations
    double          lambda;     // covariance augmentation factor
    double          chi2;       // chi-square
    double          Ochi2;      // previous chi2
    int             signum;     // index

    // gsl allocations
    covar = gsl_matrix_alloc(6, 6);
    cov = gsl_matrix_alloc(6, 6);
    alpha = gsl_matrix_alloc(6, 6);
    beta = gsl_vector_alloc(6);
    da = gsl_vector_alloc(6);
    x = gsl_vector_alloc(6);
    a = gsl_vector_alloc(6);
    aTry = gsl_vector_alloc(6);
    indx = gsl_permutation_alloc(6);

    // convert parameters of single dipole to vector
    FIELD2Vector(Dipole, a);
    chi2 = DipCoeff(Xopm, Sensor, noise, a, alpha, beta);  // solve goodness of fit (chi2) for initial parameters
    Ochi2 = chi2;

    // iterate using perturbed parameters until lambda is tiny
    lambda = ILAMBDA;
    do {

        // alter linearized fitting matrix by augmenting diagonal elements
        gsl_matrix_memcpy(covar, alpha);            // alpha --> covar
        gsl_matrix_mul_diag(covar, 1. + lambda);    // augment diagonal
        gsl_vector_memcpy(da, beta);                // beta --> da

        // matrix solution for perturbed trial vector
        gsl_matrix_memcpy(cov, covar);              // covar --> cov (cov is destroyed by LU decomposition)
        gsl_linalg_LU_decomp(cov, indx, &signum);
        gsl_linalg_LU_solve(cov, indx, x, da);

        // aTry = a + x
        gsl_vector_memcpy(aTry, a);                 // a --> aTry
        gsl_vector_add(aTry, x);                    // aTry += x

        // did the trial succeed?
        chi2 = DipCoeff(Xopm, Sensor, noise, aTry, cov, da);    // test trial parameters
        if(chi2 < Ochi2) {                          // success -- accept new solution
            lambda /= DLAMBDA;
            Ochi2 = chi2;
            gsl_matrix_memcpy(alpha, covar);        // covar --> alpha
            gsl_vector_memcpy(beta, da);            // da --> beta
            gsl_vector_memcpy(a, aTry);             // aTry --> a
        } else {                                    // failure -- increase 'lambda' & return
            lambda *= DLAMBDA;
        }
    } while(lambda > 1.0e-12);

    // free gsl allocations
    gsl_matrix_free(covar);
    gsl_matrix_free(cov);
    gsl_matrix_free(alpha);
    gsl_vector_free(beta);
    gsl_vector_free(da);
    gsl_vector_free(x);
    gsl_vector_free(a);
    gsl_vector_free(aTry);
    gsl_permutation_free(indx);

    // return sensor parameters in type FIELD
    Vector2FIELD(a, Dipole);

    // return goodness of fit
    return(chi2);
}


double  DipCoeff(                   // return chi-square
    gsl_vector  *Xopm,              // OPM sensor measurements for a single magnetic dipole - N_OPM x 1 (input)
    FIELD       *Sensor,            // Sensor[N_OPM] - array of OPM sensors (input)
    double      *noise,             // measured noise (sigma^2) for each OPM sensor (input)
    gsl_vector  *a,                 // linear dipole parameter vector (input)
    gsl_matrix  *alpha,
    gsl_vector  *beta
)
{
    FIELD       Dipole;             // test dipole parameters
    gsl_vector  *aDelta;            // test parameters - 6x1
    double      dyda[6];
    double      Xp[N_OPM];          // predicted dipole measurements value for each OPM sensor
    double      Bp;                 // positive delta field
    double      Bn;                 // negative delta field
    double      dy;
    double      wt;
    double      chi2;               // chi-square
    double      tmp;
    double      delta[] = {         // perturbations for position and orientation
        1.0e-06, 1.0e-06, 1.0e-06,  // 0.001 mm position delta
        1.0e-12, 1.0e-12, 1.0e-12   // 0.001 nA-m magnitude delta
    };
    int         i;
    int         j;
    int         s;

    // initialize deltas
    aDelta = gsl_vector_alloc(6);
    for(i=0; i<3; i++)
        delta[i] = 1.0e-06;     // 0.001 mm position delta
    for(i=3; i<6; i++)
        delta[i] = 1.0e-12;     // 0.001 nA-m magnitude delta
    
	// initialize (symmetric) alpha & beta
	gsl_matrix_set_zero(alpha);
	gsl_vector_set_zero(beta);

    // sum over all magnetic sensor locations
    Vector2FIELD(a, &Dipole);
    for(s=0, chi2=0.; s<N_OPM; s++) {

        // compute difference between predicted & measured field & accumulate chi2
        Xp[s] = MtoB(Dipole, Sensor[s]);
        dy = gsl_vector_get(Xopm, s) - Xp[s];
        chi2 += dy * dy / noise[s];

        // compute dy/da
        for(i=0; i<6; i++) {

            // compute positive delta field
            for(j=0; j<6; j++)
                if(i == j)
                    gsl_vector_set(aDelta, j, gsl_vector_get(a, j) + delta[j]);
                else
                    gsl_vector_set(aDelta, j, gsl_vector_get(a, j));
            Vector2FIELD(aDelta, &Dipole);
            Bp = MtoB(Dipole, Sensor[s]);

            // compute negative delta field
            for(j=0; j<6; j++)
                if(i == j)
                    gsl_vector_set(aDelta, j, gsl_vector_get(a, j) - delta[j]);
                else
                    gsl_vector_set(aDelta, j, gsl_vector_get(a, j));
            Vector2FIELD(aDelta, &Dipole);
            Bn = MtoB(Dipole, Sensor[s]);

            // compute partial derivative
            dyda[i] = (Bp - Bn) / (2. * delta[i]);
        }

        // compute alpha & beta
        for(i=0; i<6; i++) {
            wt = dyda[i] / noise[s];
            for(j=0; j<6; j++) {
                tmp = gsl_matrix_get(alpha, i, j) + wt * dyda[j];
                gsl_matrix_set(alpha, i, j, tmp);
            }
            tmp = gsl_vector_get(beta, i) + dy * wt;
			gsl_vector_set(beta, i, tmp);
		}
    }
    gsl_vector_free(aDelta);

    // return chi-square
    return(chi2);
}

