// FitOPM() -- fit OPM a single sensor to measurements from an array of magnetic dipoles
//
//  This subroutine employs the well-known Marquardt-Levenberg solution.
//  Measurements by a single OPM magnetometer for a known array of magnetic dipoles
//  is used to solve for the position & sensing orientation of the sensor.
//
//  Author: Stephen E. Robinson
//          MEG Core Facility
//          NIMH
//

#include "jig.h"


double  FitOPM(
    gsl_vector  *Xdip,          // sensor measurements for all magnetic dipoles - N_DIP x 1 (input)
    FIELD       *Sensor,        // pointer to a single OPM sensor's parameters position & orientation (initial guess & outout)
    double      noise,          // measured noise (input)
    FIELD       *Dipole         // array of N_DIP magnetic dipoles (input)
)
{
    gsl_matrix      *covar;     // covar - 6x6 augmented covariance of partial derivatives
    gsl_matrix      *cov;       // cov - 6x6 covariance copy
    gsl_matrix      *alpha;     // alpha - 6x6 partial derivatives
    gsl_vector      *beta;      // beta - 6x1 vector
    gsl_vector      *da;        // input vector
    gsl_vector      *x;         // solution vector
    gsl_vector      *a;         // linear OPM parameters
    gsl_vector      *aTry;      // trial parameters
	gsl_permutation	*indx;      // LU decomposition permutations
    double          lambda;     // covariance augmentation factor
    double          chi2;       // chi-square
    double          Ochi2;      // previous chi2
    int             signum;     // index

    // gsl allocations
    covar = gsl_matrix_alloc(6, 6);
    cov = gsl_matrix_alloc(6, 6);
    alpha = gsl_matrix_alloc(6, 6);
    beta = gsl_vector_alloc(6);
    da = gsl_vector_alloc(6);
    x = gsl_vector_alloc(6);
    a = gsl_vector_alloc(6);
    aTry = gsl_vector_alloc(6);
    indx = gsl_permutation_alloc(6);

    // convert OPM sensor parameters to vector
    FIELD2Vector(Sensor, a);
    chi2 = OPMCoeff(Xdip, Sensor, noise, Dipole, a, alpha, beta);   // solve goodness of fit (chi2) for initial parameters
    Ochi2 = chi2;

    // iterate using perturbed parameters until lambda is tiny
    lambda = ILAMBDA;
    do {

        // alter linearized fitting matrix by augmenting diagonal elements
        gsl_matrix_memcpy(covar, alpha);            // alpha --> covar
        gsl_matrix_mul_diag(covar, 1. + lambda);    // augment diagonal
        gsl_vector_memcpy(da, beta);                // beta --> da

        // matrix solution for perturbed trial vector
        gsl_matrix_memcpy(cov, covar);              // covar --> cov (cov is destroyed by LU decomposition)
        gsl_linalg_LU_decomp(cov, indx, &signum);
        gsl_linalg_LU_solve(cov, indx, x, da);

        // aTry = a + x
        gsl_vector_memcpy(aTry, a);                 // a --> aTry
        gsl_vector_add(aTry, x);                    // aTry += x

        // did the trial succeed?
        chi2 = OPMCoeff(Xdip, Sensor, noise, Dipole, aTry, covar, da);  // test trial parameters
        if(chi2 < Ochi2) {                        // success -- accept new solution
            lambda /= DLAMBDA;
            Ochi2 = chi2;
            gsl_matrix_memcpy(alpha, covar);        // covar --> alpha
            gsl_vector_memcpy(beta, da);            // da --> beta
            gsl_vector_memcpy(a, aTry);             // aTry --> a
        } else {                                    // failure -- increase 'lambda' & return
            lambda *= DLAMBDA;
        }
    } while(lambda > 1.0e-12);

    // free gsl allocations
    gsl_matrix_free(covar);
    gsl_matrix_free(cov);
    gsl_matrix_free(alpha);
    gsl_vector_free(beta);
    gsl_vector_free(da);
    gsl_vector_free(x);
    gsl_vector_free(a);
    gsl_vector_free(aTry);
    gsl_permutation_free(indx);

    // return fitted sensor parameters in type FIELD
    Vector2FIELD(a, Sensor);

    // return goodness of fit
    return(chi2);
}


double  OPMCoeff(                   // return chi-square
    gsl_vector  *Xdip,              // dipole measurement vector for this OPM sensor (input)
    FIELD       *Sensor,            // pointer to single test OPM magnetometer sensor parameters (input)
    double      noise,              // measured noise (sigma^2) for this OPM sensor (input)
    FIELD       *Dipole,            // array of magnetic dipole positions & orientations (input)
    gsl_vector  *a,                 // OPM parameter vector (input)
    gsl_matrix  *alpha,
    gsl_vector  *beta
)
{
    FIELD       aSensor;            // test OPM magnetometer vector
    gsl_vector  *aDelta;            // test linear parameters - 6x1
    double      dyda[6];            // partial derivatives
    double      Xp[N_DIP];          // predicted OPM signal value for each magnetic dipole
    double      Bp;                 // positive delta field
    double      Bn;                 // negative delta field
    double      dy;                 // difference between measured & predicted
    double      chi2;               // chi-square
    double      wt;                 // weight
    double      tmp;
    double      delta[] = {         // perturbations for position and orientation
        1.0e-06, 1.0e-06, 1.0e-06,  // 0.001 mm position delta
        1.0e-12, 1.0e-12, 1.0e-12   // 0.001 nA-m magnitude delta
    };
    int         d;                  // dipole index
    int         i;
    int         j;

    // initialize aDelta & deltas
    aDelta = gsl_vector_alloc(6);
    for(i=0; i<3; i++)
        delta[i] = 1.0e-06;     // 0.001 mm position delta
    for(i=3; i<6; i++)
        delta[i] = 1.0e-12;     // 0.001 nA-m magnitude delta
    
	// initialize (symmetric) alpha & beta
	gsl_matrix_set_zero(alpha);
	gsl_vector_set_zero(beta);

    // compute test sensor field for _all_ magnetic dipoles
    for(d=0; d<N_DIP; d++)
        Xp[d] = MtoB(Dipole[d], &Sensor);

    // sum over all magnetic dipole locations
    Vector2FIELD(a, Sensor);
    for(d=0, chi2=0.; d<N_DIP; d++) {

        // compute difference between predicted & measured field & accumulate chi2
        dy = gsl_vector_get(Xdip, d) - Xp[d];
        chi2 += dy * dy / noise;

        // compute dy/da
        for(i=0; i<6; i++) {

            // compute positive delta field
            for(j=0; j<6; j++)
                if(i == j)
                    gsl_vector_set(aDelta, j, gsl_vector_get(a, j) + delta[j]);
                else
                    gsl_vector_set(aDelta, j, gsl_vector_get(a, j));
            Vector2FIELD(aDelta, &aSensor);
            Bp = MtoB(Dipole[d], aSensor);

            // compute negative delta field
            for(j=0; j<6; j++)
                if(i == j)
                    gsl_vector_set(aDelta, j, gsl_vector_get(a, j) - delta[j]);
                else
                    gsl_vector_set(aDelta, j, gsl_vector_get(a, j));
            Vector2FIELD(aDelta, &aSensor);
            Bn = MtoB(Dipole[d], aSensor);

            // compute partial derivative
            dyda[i] = (Bp - Bn) / (2. * delta[i]);
        }

        // compute alpha & beta
        for(i=0; i<6; i++) {
            wt = dyda[i] / noise;
            for(j=0; j<6; j++) {
                tmp = gsl_matrix_get(alpha, i, j) + wt * dyda[j];
                gsl_matrix_set(alpha, i, j, tmp);
            }
            tmp = gsl_vector_get(beta, i) + dy * wt;
			gsl_vector_set(beta, i, tmp);
		}
    }

    gsl_vector_free(aDelta);

    // return chi-square
    return(chi2);
}
