// OPMcal -- calibration procedure for OPM sensor array
//
//  Let us first assume that the calibration jig with its magnetic dipoles is dimensionally accurate.
//  This implies that the sensors, when calibrated, must localize each magnetic dipole such that
//  deviations from their "ground truth" positions is minimized. Thus, the convergence criterion is
//  that chi^2 of the measured magnetic dipole positions vs their true positions is minimized.
//
//  Given a table/file with the true dipole positions relative to the jig origin, we first fit the
//  position & orientation for each OPM sensor, using the measurments from the array of magnetic
//  dipoles. Next, we compute position (discarding the orientation) of each magnetic dipole using the
//  measurements from the array of OPM sensors. We use chi^2 as a measure of the deviations of the
//  magnetic dipoles from their true locations. Iterating between fitting sensors and fitting dipoles
//  reduces the chi^2 error.
//
//  Starting with the nominal sensor calibrations and exact "true" magnetic dipole positions, the
//  steps are:
//
//      1) Fit each OPM sensor (position & orientation) to the array of [37] magnetic dipoles
//      2) Fit the positions of each magnetic dipole to the array of OPM sensors
//      3) Compute chi^2 of error between "ground truth" dipole positions & computed positions
//      4) repeat steps 1-3 until chi^2 is minimized
//      5) Compute sensor gain from measured dipole moment (current through dipoles)
//      6) Save OPM sensor relative locations, orientations, & gain
//
//  Variables:
//      gsl_matrix  *Xm - N_OPMxN_DIP - measured magnetic field for all sensors & all dipoles
//      FIELD   Sensor[N_OPM] - array of OPM sensors
//      FIELD   Dipole[N_DIP] - array of magnetic dipoles
//      double  noise[N_OPM] - noise of each sensor
//
//  This calibration routine is agnostic to the OPM data format. Instead, it is assumed that
//  the OPM data are collected after determining the optimal frequency for driving the dipoles.
//  The OPM array rms signals for each magnetic dipole, rms noise in the absence of a dipole
//  drive, & rms current are determined for the calibrator frequency (FFT or correlation).
//  The resulting values are stored as text in the following files.
//
//  Files:
//      "CalData" - N_OPM x N_DIP list of sensor rms signal at frequency of calibrator  (input)
//      "NoiseData" - N_OPM list of sensor rms noise at frequncy of calibrator          (input)
//      "CurrentData" - N_DIP list of rms calibration currents                          (input)
//      "SensorCalibrations" - N_OPM list of sensor positions, orientations, & gains    (output)
//
//  Author: Stephen E. Robinson
//          MEG Core Facility
//          NIMH
//

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <samlib.h>
#include <DataFiles.h>
#include <samutil.h>
#include "jig.h"
#include "truth.h"                  // calibration jig locations & starting sensor values


int	main(
	int		argc,
	char	**argv)
{
    FIELD       Dipole[N_DIP];      // magnetic dipole positions & orientation vectors (with moment)
    FIELD       Sensor[N_OPM];      // OPM sensor positions & orientation vectors
    gsl_matrix  *Xm;                // measured field of array of magnetic dipoles for each OPM sensor - N_OPM x N_DIP
    gsl_vector  *Xopm;              // measured field vector for array of OPM sensors for a single dipole
    gsl_vector  *Xdip;              // measured field vector for array of magnetic dipoles for a single OPM sensor
    double      gain[N_OPM];        // OPM gain values T/bit
    double      current[N_DIP];     // dipole current values
    double      noise[N_OPM];       // OPM noise values
    double      moment;             // magnetic dipole moment
    double      measured;           // measured moment
    double      predicted;          // predicted moment
    double      chi2;               // chi-square
    double      old_chi2;           // previous chi^2
    double      r;                  // distance
    double      r2;                 // r^2
    double      gof;                // goodness-of-fit
    double      tmp;                // a nameless variable
    int         d;                  // dipole index
    int         s;                  // sensor index
    int         v;                  // vector index
    int         pass;               // iteration number
    char        fpath[256];         // file path
    FILE        *fp;                // file pointer

    // allocate gsl objects
    Xm = gsl_matrix_alloc(N_OPM, N_DIP);
    Xopm = gsl_vector_alloc(N_OPM);
    Xdip = gsl_vector_alloc(N_DIP);

    // (1) establish initial parameters for OPM sensors & magnetic dipoles
    for(d=0; d<N_DIP; d++)
        for(v=X_; v<=Z_; v++) {
            Dipole[d].p[v] = 1000. * Jig[d][v];     // convert mm to m (from "truth.h")
            Dipole[d].v[v] = 1.0e-12;               // nominal starting value
        }
    for(s=0; s<N_OPM; s++) {
        for(v=X_; v<=Z_; v++)
            Sensor[s].p[v] = 1000. * Mag[s][v];     // convert mm to m (from "truth.h")
        for(v=3; v<=6; v++)
            Sensor[s].v[v] = Mag[s][v];             // nominal orientation from sensor fixture
    }

    // (2) read OPM sensor data
    sprintf(fpath, "CalData");
    if((fp = fopen(fpath, "r")) == NULL)
        Cleanup("can't open calibration data file");

    for(s=0; s<N_OPM; s++)
        for(d=0; d<N_DIP; d++) {
            fscanf(fp, "%lf", &tmp);
            gsl_matrix_set(Xm, s, d, tmp);
        }
    fclose(fp);

    // (3) read OPM noise data
    sprintf(fpath, "NoiseData");
    if((fp = fopen(fpath, "r")) == NULL)
        Cleanup("can't open sensor noise data file");
    for(s=0; s<N_OPM; s++)
        fscanf(fp, "%lf", &noise[s]);
    fclose(fp);

    // (4) read magnetic dipole current
    sprintf(fpath, "CurrentData");
    if((fp = fopen(fpath, "r")) == NULL)
        Cleanup("can't open magnetic dipole current data file");
    for(d=0; d<N_DIP; d++)
        fscanf(fp, "%lf", &current[d]);
    fclose(fp);

    // (5) iterate calibration steps 5 & 6 until results are stable
    chi2 = HUGE;
    pass = 1;
    do {

        // save chi2
        old_chi2 = chi2;

        // (6) determine parameters of each OPM sensor using jig frame origin & _exact_ relationship of each magnetic dipole
        printf("fitting each OPM sensor to array of magnetic dipoles");
        for(s=0; s<N_OPM; s++) {
            gsl_matrix_get_row(Xdip, Xm, s);
            gof = FitOPM(Xdip, &Sensor[s], noise[s], Dipole);    // update sensor position & orientation
            printf(".");
            fflush(stdout);
        }

        // (7) using current OPM sensor parameters, compute chi^2 of deviation of magnetic dipole locations from their true positions
        printf(" - done\n");
        printf("fitting each magnetic dipole to array of OPM sensors");
        for(d=0, chi2=0.; d<N_DIP; d++) {
            gsl_matrix_get_col(Xopm, Xm, d);
            gof = FitDip(Xopm, Sensor, noise, &Dipole[d]);  // fit one dipole to the sensor array
            for(v=X_; v<=Z_; v++) {
                tmp = Dipole[d].p[v] - Jig[d][v];           // compute difference between measured & "ground truth" coordinates
                chi2 += tmp * tmp / Jig[d][v];              // accumulate chi^2
            }
            printf(".");
            fflush(stdout);
        }

        // show pass results
        printf(" - done\n");
        printf("pass %d, chi^2=%e\n", pass, chi2);
        fflush(stdout);
        pass++;

    } while(chi2 < old_chi2);

    // (8) normalize all sensor orientation vectors
    for(s=0; s<N_OPM; s++) {
        for(v=X_, r2=0.; v<=Z_; v++)
            r2 += Sensor[s].v[v] * Sensor[s].v[v];
        r = sqrt(r2);
        for(v=X_; v<=Z_; v++)
            Sensor[s].v[v] /= r;
    }

    // (9) fit sensor gains using measured currents for each dipole
    for(d=0; d<N_DIP; d++) {
        moment = AM2 * current[d];
        for(v=X_; v<=Z_; v++)
            Dipole[d].v[v] *= moment;
    }
    for(s=0; s<N_OPM; s++) {
        for(d=0, gain[s]=0.; d<N_DIP; d++) {
            predicted = MtoB(&Dipole[d], &Sensor[s]);
            measured = gsl_matrix_get(Xm, s, d);     // measured may be inaccurate due to gain error
            gain[s] += predicted;
        }
        gain[s] /= (double)N_DIP;
    }

    // (10) report & save calibration results
    sprintf(fpath, "SensorCalibrations");
    if((fp = fopen(fpath, "w")) == NULL);
        Cleanup("can't open sensor calibration file");
    for(s=0; s<N_OPM; s++)
        fprintf(fp, "%f%f%f%f%f%f%e", Sensor[s].p[X_], Sensor[s].p[Y_], Sensor[s].p[Z_], Sensor[s].v[X_], Sensor[s].v[Y_], Sensor[s].v[Z_], gain[s]);
    fclose(fp);

    // free gsl objects
    gsl_matrix_free(Xm);
    gsl_vector_free(Xopm);
    gsl_vector_free(Xdip);
}
