// OPMcal -- calibration procedure for OPM sensor array
//
//  Let us first assume that the calibration jig with its magnetic dipoles is dimensionally accurate.
//  This implies that the sensors, when calibrated, must localize each magnetic dipole such that
//  deviations from their "ground truth" positions is minimized. Thus, the convergence criterion is
//  that chi^2 of the measured magnetic dipole positions vs their true positions is minimized.
//
//  Given a table/file with the true dipole positions relative to the jig origin, we first fit the
//  position & orientation for each OPM sensor, using the measurments from the array of magnetic
//  dipoles. Next, we compute position (discarding the orientation) of each magnetic dipole using the
//  measurements from the array of OPM sensors. We use chi^2 as a measure of the deviations of the
//  magnetic dipoles from their true locations. Iterating between fitting sensors and fitting dipoles
//  reduces the chi^2 error.
//
//  Starting with the nominal sensor calibrations and exact "true" magnetic dipole positions, the
//  steps are:
//
//      1) Fit each OPM sensor (position & orientation) to the array of [37] magnetic dipoles
//      2) Fit the positions of each magnetic dipole to the array of OPM sensors
//      3) Compute chi^2 of error between "ground truth" dipole positions & computed positions
//      4) repeat steps 1-3 until chi^2 is minimized
//      5) Compute sensor gain from measured dipole current through dipole coils
//      6) Save OPM sensor relative locations, orientations, & gain
//
//  Variables:
//      gsl_matrix  *Measured - N_OPM x N_DIP - measured rms calibrator measurements
//      gsl_matrix  *Variance - N_OPM x N_DIP - measured sensor noise variance
//      FIELD       Sensor[N_OPM] - array of OPM sensors
//      FIELD       Dipole[N_DIP] - array of magnetic dipole coils
//
//  This calibration routine is agnostic to the OPM data format. Instead, it is assumed that
//  the OPM data are collected after determining the optimal frequency for driving the dipoles.
//  The OPM array rms signals for each magnetic dipole, rms noise in the absence of a dipole
//  drive, & rms current are determined for the calibrator frequency (FFT or correlation).
//  The resulting values are stored as text in the following files.
//
//  Files:
//      "CalData" - N_OPM x N_DIP list of sensor rms signal at frequency of calibrator  (input)
//      "NoiseData" - N_OPM x N_DIP list of sensor rms noise at frequency of calibrator (input)
//      "VoltageData" - N_DIP list of rms calibration voltage                           (input)
//      "SensorCalibrations" - N_OPM list of sensor positions, orientations, & gains    (output)
//
//  Instructions for creating required files:
//      Execute "OPMstats -t 2" on all calibration datasets
//      Execute "OPMstats -t 2 -i" on at least one dataset to generate the sensor list
//      "cat *,Ayy.txt > VoltageData.txt"; insert first line with number of dipole coils, e.g., 37
//      "cat *,Axx.txt > CalData.txt"; insert first line with number of sensors & number of dipole coils, e.g., 19 37
//      "cat *,AxxVar.txt > NoiseData"; insert first line with number of sensors & number of dipole coils, e.g., 19 37
//      
//  Author: Stephen E. Robinson
//          MEG Core Facility
//          NIMH
//

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <samlib.h>
#include <DataFiles.h>
#include <samutil.h>
#include <jig.h>
#include <truth.h>                  // calibration jig locations & starting sensor values

#define SIGN(a) ( a/fabs(a) )

#define DEBUG   TRUE

int	main(
	int		argc,
	char	**argv)
{
    FIELD       *Dipole;            // Dipole[D] -- computed magnetic dipole positions & orientation vectors
    FIELD       *Coil;              // Coil[D] -- magnetic dipole positions & orientation vectors
    FIELD       *Sensor;            // Sensor[S] -- OPM sensor positions & orientation vectors
    FIELD       *OldSensor;         // OldSensor[S] -- previous sensor geometry
    gsl_matrix  *Measured;          // measured field of array of magnetic dipoles for each OPM sensor - N_OPM x N_DIP
    gsl_matrix  *Variance;          // sensor noise variance - N_OPM x N_DIP
    double      *Current;           // Current[N_DIP] -- rms dipole coil current (A)
    double      *Gain;              // Gain[S] -- OPM gain values
    double      *OldGain;           // previous OPM gains
    double      **True;             // True[D][D] -- distances between dipoles
    double      measured;           // measured rms field
    double      predicted;          // predicted rms field
    double      chi2;               // chi-square
    double      old_chi2;           // previous chi^2
    double      r;                  // distance
    double      r2;                 // r^2
    double      gof;                // goodness-of-fit
    double      adc;                // calibrator rms voltage
    double      tmp;                // a nameless variable
    int         d;                  // dipole index
    int         s;                  // sensor index
    int         ss;                 // alternate sensor index
    int         i;                  // i-index
    int         j;                  // j-index
    int         v;                  // vector index
    int         D;                  // number of dipoles in files
    int         S;                  // number of sensors in files
    int         pass;               // iteration number
    char        fpath[256];         // file path
    FILE        *fp;                // file pointer

    // (1) read number of sensors & dipoles
    printf("reading dimensions from 'CalData.txt");
    fflush(stdout);
    sprintf(fpath, "CalData.txt");
    if ((fp = fopen(fpath, "r")) == NULL)
        Cleanup("can't open 'CalData.txt' file");
    fscanf(fp, "%d%d", &S, &D);     // read actual numbers from calibration data
    if (D != N_DIP)                 // sanity check
        Cleanup("number of dipoles in 'CalData.txt' disagrees with 'truth.h'");

    // (2) allocate arrays
    printf(" - done\n");
    printf("allocating arrays");
    fflush(stdout);
    Measured = gsl_matrix_alloc(S, D);
    Variance = gsl_matrix_alloc(S, D);
    Dipole = new_arrayE(FIELD, D, "Dipole[D]");
    Coil = new_arrayE(FIELD, D, "Coil[D]");
    Sensor = new_arrayE(FIELD, S, "Sensor[S]");
    OldSensor = new_arrayE(FIELD, S, "OldSensor[S]");
    Gain = new_arrayE(double, S, "Gain[S]");
    OldGain = new_arrayE(double, S, "OldGain[S]");
    Current = new_arrayE(double, D, "Current[D]");
    True = new_matrixE(D, D, "True[D][D]");

    // (3) compute distance matrix for dipole true coordinates
    printf(" - done\n");
    printf("computing true inter-coil distance matrix\n");
    fflush(stdout);
    for (i=0; i<D; i++)
        for (j=0; j<D; j++) {
            for (v=X_, r2=0.; v<=Z_; v++) {
                tmp = 1.0e-03 * (Jig[i][v] - Jig[j][v]);    // jig coordinates were mm
                r2 += tmp * tmp;
            }
            r = sqrt(r2);
            True[i][j] = r;
        }

    // (4) read calibration data
    printf(" - done\n"); 
    printf("reading 'CalData.txt'");
    fflush(stdout);
    for (d=0; d<D; d++)
        for (s=0; s<S; s++) {
            fscanf(fp, "%d%lf", &ss, &tmp);
            gsl_matrix_set(Measured, s, d, tmp);
        }
    fclose(fp);

    // (5) read OPM noise data
    printf(" - done\n");
    printf("reading 'NoiseData.txt'");
    fflush(stdout);
    sprintf(fpath, "NoiseData.txt");    // NoiseData.txt lists rms noise
    if((fp = fopen(fpath, "r")) == NULL)
        Cleanup("can't open sensor noise data file");
    fscanf(fp, "%d%d", &S, &D);         // read actual numbers from noise data
    if (D != N_DIP)                     // sanity check
        Cleanup("number of dipoles in 'NoiseData.txt' disagrees with 'truth.h'");
    for(d=0; d<D; d++)
        for (s=0; s<S; s++) {
            fscanf(fp, "%d%lf", &ss, &tmp);
            gsl_matrix_set(Variance, s, d, tmp);    // 'NoiseData' values are based on '*,AxxVar.txt'
        }
    fclose(fp);

    // (6) read calibrator output voltage
    printf(" - done\n");
    printf("reading 'VoltageData.txt'");
    fflush(stdout);
    sprintf(fpath, "VoltageData.txt");
    if ((fp = fopen(fpath, "r")) == NULL)
        Cleanup("can't open calibrator voltage data file");
    fscanf(fp, "%d", &D);
    if (D != N_DIP)
        Cleanup("number of dipole in 'VoltageData.txt' disagrees with 'truth.h'");
    for(d=0; d<D; d++) {
        fscanf(fp, "%lf%lf", &adc, &tmp);   // OPMstats reports V rms
        Current[d] = Sign[d] * V2A * adc;     // convert volts to current
    }
    fclose(fp);

    // (7) establish initial parameters for OPM sensors & magnetic dipoles
    printf(" - done\n");
    printf("setting initial parameters from 'truth.h'");
    fflush(stdout);
    if ((fp = fopen("OPMlist", "r")) == NULL)
        Cleanup("can't open 'OPMlist' for read");
    fscanf(fp, "%d\n", &ss);
    if (ss != S)
        Cleanup("number of OPMs in list disagrees with data");
    for (s=0; s<S; s++) {
        fscanf(fp, "%d\n", &ss);
#ifdef DEBUG
        printf("%d in data corresponds with Mag[%d]\n", s, ss);
        fflush(stdout);
#endif
        for (v=X_; v<=Z_; v++) {
            Sensor[s].p[v] = OldSensor[s].p[v] = 0.001 * Mag[ss][v];    // convert mm to m (from "truth.h")
            Sensor[s].v[v] = OldSensor[s].v[v] = Mag[ss][v+3];          // nominal orientation from sensor fixture
            
        }
    }
    fclose(fp);
    for (d=0; d<D; d++)                             // 'Coil' has the ground truth
        for (v=X_; v<=Z_; v++) {
            Coil[d].p[v] = 0.001 * Jig[d][v];
            Coil[d].v[v] = Jig[d][v+3];
        }

    // don't forget to initialize 'Dipole', too!
    for (d=0; d<D; d++)
        for (v=X_; v<=Z_; v++) {
            Dipole[d].p[v] = Coil[d].p[v];
            Dipole[d].v[v] = Coil[d].v[v];
        }

    // initialize 'OldGain'
    for (s=0; s<S; s++)
        OldGain[s] = 1.;
    
    // (8) iterate calibration steps 5 & 6 until results are stable
    printf(" - done\n");
    printf("iterating calibrations:\n");
    fflush(stdout);
    chi2 = HUGE;
    pass = 1;
    do {

        // save chi2
        old_chi2 = chi2;

        // (8a) determine parameters of each OPM sensor using jig frame origin & _exact_ relationship of each magnetic dipole
        printf("fitting each OPM sensor to array of magnetic dipoles");
        for (s=0; s<S; s++) {
            gof = FitOPM(s, Measured, Variance, Current, &Gain[s], Sensor, Coil);   // fit one sensor position, orientation
#ifdef DEBUG
            if (s == 0)
                printf("\n");
            printf("Sensor%d: gof=%e\t%f\t%f\t%f\t%f\t%f\t%f\t%f\n", s, gof,
              1000.*Sensor[s].p[X_], 1000.*Sensor[s].p[Y_], 1000.*Sensor[s].p[Z_], Sensor[s].v[X_], Sensor[s].v[Y_], Sensor[s].v[Z_], Gain[s]);
#else
            printf(".");
            fflush(stdout);
#endif
        }

        // (8b) using current OPM sensor parameters, compute chi^2 of deviation of magnetic dipole locations from their true positions
        printf("\nfitting each magnetic dipole to array of OPM sensors");
        for (d=0; d<D; d++) {
            gof = FitDip(d, Measured, Variance, Current, Sensor, Dipole); // fit one dipole to the sensor array

            // for the next iteration, restore true dipole coil position, while allowing coil orientation to change
            for (v=X_; v<=Z_; v++) {
                Coil[d].p[v] = 0.001 * Jig[d][v];
                Coil[d].v[v] = Dipole[d].v[v];
            }
#ifdef DEBUG
            if (d == 0)
                printf("\n");
            printf("Dipole%d: gof=%e\t%f\t%f\t%f\t%f\t%f\t%f\n", d, gof,
              1000.*Coil[d].p[X_], 1000.*Coil[d].p[Y_], 1000.*Coil[d].p[Z_], Coil[d].v[X_], Coil[d].v[Y_], Coil[d].v[Z_]);
#else
            printf(".");
            fflush(stdout);
#endif
        }

        // (8c) compute chi^2 using relative distances
        printf("\n");
        printf("computing error of fit to true inter-coil distances");
        fflush(stdout);
        for (i=0, chi2=0.; i<D; i++)
            for (j=0; j<D; j++)
                if (i != j) {
                    for (v=X_, r2=0.; v<Z_; v++) {
                        tmp = Dipole[i].p[v] - Dipole[j].p[v];
                        r2 += tmp * tmp;
                    }
                    r = sqrt(r2);
                    tmp = r - True[i][j];
                    chi2 += (tmp * tmp) / True[i][j];
                }

        // (8d) save the sensor geometry if chi^2 improves
        if (chi2 < old_chi2)
            for (s=0; s<S; s++) {
                for (v=X_; v<=Z_; v++) {
                    OldSensor[s].p[v] = Sensor[s].p[v];
                    OldSensor[s].v[v] = Sensor[s].v[v];
                }
                OldGain[s] = Gain[s];
            }

        // show pass results
        printf(" - done\n");
        printf("pass %d, reduced chi^2=%e\n", pass, chi2 / (double)D);
        fflush(stdout);
        pass++;

    } while (chi2 < old_chi2);

    // (9) report sensor gains
    if ((fp = fopen("SensorGain.txt", "w")) == NULL)
        Cleanup("can't open 'SensorGain.txt' for write");
    for (s=0; s<S; s++) {
        fprintf(fp, "%d\t%e\n", s, OldGain[s]);
        fflush(fp);
    }
    fclose(fp);

    // (10) report & save calibration results
    printf(" - done\n");
    printf("writing sensor calibrations\n");
    fflush(stdout);
    if ((fp = fopen("SensorCalibration.txt", "w")) == NULL)
        Cleanup("can't open sensor calibration file");
    for (s=0; s<S; s++)
        fprintf(fp, "%d\t%f\t%f\t%f\t%f\t%f\t%f\t%e\n", s, 1000.*OldSensor[s].p[X_], 1000.*OldSensor[s].p[Y_], 1000.*OldSensor[s].p[Z_],
          OldSensor[s].v[X_], OldSensor[s].v[Y_], OldSensor[s].v[Z_], OldGain[s]);
    fclose(fp);

    // free gsl objects
    gsl_matrix_free(Measured);
    gsl_matrix_free(Variance);
    free(Dipole);
    free(Coil);
    free(Sensor);
    free(OldSensor);
    free(Gain);
    free(OldGain);
    free(Current);

    printf("OPMcal done!\n");
    fflush(stdout);
}
