// OPM_Stats -- reads a dataset and computes the crosspower & gain
//  between the reference input & each of the OPM sensors. The data
//  are parsed into overlapping segments of 'SegTime' duration.
//
// The spectral values, averaged over the number of segments, and
//  normalized by the segment size are:
//      Axx -- autopower of OPM sensor (real)
//      Ayy -- autopower of reference (real)
//      Cxy -- crosspower (complex)
//
// The phase of the OPM sensor relative to the reference is computed
//  from Cxy (complex). OPM rms sensor output (at the calibrator
//  frequency) is sqrt(Axx). Coherence (Rxy) between reference and
//  each OPM sensor is (Cxy * Cxy*) / (Axx * Ayy). The "Gain" (more
//  properly, the transfer function from reference to each OPM) is
//  Axx / |Cxy|. All values are maintained in uscaled SI units.
//
// Usage:  OPM_Stats
//          -r dataset name
//          -t window size (s)
//          -i generate sensor index
//          -v verbose
//
// Author: Stephen E. Robinson
//          MEG Core Facility
//          NIMH
//

#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <fcntl.h>
#include <ctype.h>
#include <string.h>
#include <strings.h>
#include <siglib.h>
#include <complex.h>
#include <fftw3.h>
#include <samlib.h>
#include <DataFiles.h>
#include <samutil.h>
#include <jig.h>
#include <truth.h>

// req'd constants in jig.h
#define CX0     (0.+0.*I)       // complex zero

// #define EBUG    TRUE


int main(
    int         argc,
    char        **argv
) {

    // variable declaration
    HeaderInfo      Header;     // header information data
    ChannelInfo     *Channel;   // Channel[M] -- channel information data
    EpochInfo       *Epoch;     // Epoch[E] -- epoch information data
    fftw_plan       PsFwd;      // plan primary sensor real-to-complex fft
    fftw_plan       RsFwd;      // plan for reference sensor real-to-complex fft
    double          **Data;     // Data[S][T] -- OPM data array
    double          *In;        // In[T] -- single channel data
    double          *Seg;       // Seg[TW] -- temporary data segment
    double          *Hann;      // Hann[TW] -- Hanning window
    double          *x;         // x[TW] -- real sensor segment
    double          *y;         // y[TW] -- real reference segment
    fftw_complex    *X;         // X[TW] -- complex sensor segment spectrum
    fftw_complex    *Y;         // Y[TW] -- complex reference segment spectrum
    fftw_complex    Xo;         // real & imaginary from fft
    fftw_complex    Yo;         // real & imaginary from fft
    fftw_complex    *Cxy;       // Cxy[M] -- sensor-reference crosspowers (two complex numbers yields a complex number)
    fftw_complex    **SegXo;    // SegXo[M][NumSegs] -- sensor segment spectral line
    fftw_complex    *SegYo;     // SegYo[NumSegs] -- reference segment spectral line
    double          *Axx;       // Axx[M] -- sensor autopower
    double          *AxxVar;    // AxxVar[M] -- sensor variance
    double          *sign;      // polarity of OPM rms signal
    double          Ayy;        // reference autopower
    double          Rxy;        // coherence
    double          Gain;       // reference to sensor transfer function
    double          Phase;      // sensor phase relative to ADC
    double          SegTime;    // segment duration
    double          time;       // dataset duration
    double          scale;      // conversion to Tesla
    double          freq;       // frequency (Hz)
    double          df;         // delta f (Hz)
    double          bins;       // number of spectral bins
    double          Hxx;        // harmonic autopower (& single sensor autopower
    double          max;        // maximum reference power
    double          min;        // minimum sensor power
    double          sx;         // sum of x
    double          sx2;        // sum of x^2
    double          m2;         // spectal line autopower squared
    double          stdev;      // standard deviation
    double          tmp;        // the variable that we cannot do without...
    extern char     *optarg;
    extern int      optind;
    extern int      opterr;
    int             c;          // command index
    int             h;          // harmonic index
    int             m;          // sensor index
    int             n;          // alternate sensor index
    int             s;          // selected spectral bin
    int             ss;         // alternate spectral bin
    int             f;          // search spectral bin
    int             t;          // sample index
    int             tt;         // alternate sample index
    int             w;          // window index
    int             S;          // total number of channels (including ADC)
    int             M;          // number of OPM sensors
    int             T;          // total samples
    int             TW;         // number of samples in segment
    int             Step;       // number of samples per segment step
    int             NumSegs;    // number of segments
    int             eflg = FALSE;
    int             iflg = FALSE;
    int             rflg = FALSE;
    int             tflg = FALSE;
    int             vflg = FALSE;
    char            SetName[64];
    char            OutName[64];
    FILE            *fp;
    
    opterr = 1;     // enable error reporting
    while ((c = getopt(argc, argv, "r:t:iv")) != EOF) {
        switch (c) {

            case 'r':   // dataset name
                sprintf(SetName, "%s", optarg);
                rflg = TRUE;
                break;

            case 't':   // duration of segments
                SegTime = atof(optarg);
                tflg = TRUE;
                break;

            case 'i':   // generate OPM index
                iflg = TRUE;
                break;

            case 'v':   // verbose
                vflg = TRUE;
                break;

            default:
                eflg = TRUE;
                break;
        }
    }
    if (eflg || !rflg || !tflg) {
        fprintf(stderr, "OPM_Stats\t-r <dataset name>\t%s\n", __DATE__);
        fprintf(stderr, "\t\t-t <segment duration (s)>\n");
        fprintf(stderr, "\t\t-i -- generate OPM index\n");
        fprintf(stderr, "\t\t-v -- verbose mode\n");
        exit(-1);
    }

    // announce & get data with sensor structures
	if (vflg) {
        printf("S P E C T R A L   C R O S S P O W E R  &  G A I N\n");
        printf("opening dataset");
        fflush(stdout);
    }
    GetDsInfo(SetName, &Header, &Channel, &Epoch, NULL, TRUE);
    if (Header.NumEpochs != 1)
        Cleanup("single epoch dataset req'd");
    S = Header.NumChannels;
    M = Header.NumPri + Header.NumRef;
    T = Header.MaxSamples;          // number of samples per channel per epoch
    time = (double)T / Header.SampleRate;
    TW = (int)floor(SegTime * Header.SampleRate);   // floor!
    Step = TW / 2;
    NumSegs = T / Step - 1;         // for 50% overlap
    df = 1. / SegTime;
    bins = (double)TW / 2. + 1.;    // number of spectral bins (for normalization of spectra)

    // assume channel list -- ADC reference is Channel[0]
    if (vflg) {
        printf(" - done\n");
        printf("total number of channels acquired: %d\n", Header.NumChannels);
        printf("number of primary sensors: %d\n", Header.NumPri);
        printf("number of reference sensors: %d\n", Header.NumRef);
        printf("ADC is channel 0\n");
        printf("dataset duration %5.1f seconds (%d samples) at %5.1f Hz\n", time, T, Header.SampleRate);
        printf("data will be segmented into %d %4.1f second windows (%d samples)\n", NumSegs, SegTime, TW);
        printf("spectral resolution is %3.1f Hz\n", df);
        printf("creating channel list");
        fflush(stdout);
    }

    // determine segment size & allocate memory
    Data = new_matrixE(S, T, "Data[S][T]");
    In = new_arrayE(double, T, "In[T]");
    x = new_arrayE(double, TW, "x[TW]");
    y = new_arrayE(double, TW, "y[TW]");
    X = fftw_malloc(TW * sizeof(fftw_complex));
    Y = fftw_malloc(TW * sizeof(fftw_complex));
    SegXo = (fftw_complex **)fftw_malloc(M * sizeof(fftw_complex *));
    for (m=0; m<M; m++)
        SegXo[m] = (fftw_complex *)fftw_malloc(NumSegs * sizeof(fftw_complex));
    SegYo = (fftw_complex *)fftw_malloc(NumSegs * sizeof(fftw_complex));
    Cxy = fftw_malloc(S * sizeof(fftw_complex));
    AxxVar = new_arrayE(double, M, "AxxVar[M]");
    sign = new_arrayE(double, M, "sign[M]");
    Hann = new_arrayE(double, TW, "Hann[TW]");
    Seg = new_arrayE(double, TW, "Seg[TW]");
    Axx = new_arrayE(double, M, "Axx[M]");

    // generate Hanning window
    Hanning(Hann, TW);

    // create fft plans
    PsFwd = fftw_plan_dft_r2c_1d(TW, x, X, FFTW_ESTIMATE);      // for primary sensors
    RsFwd = fftw_plan_dft_r2c_1d(TW, y, Y, FFTW_ESTIMATE);      // for ADC reference

    // read dataset
    if (vflg) {
        printf("reading data");
        fflush(stdout);
    }
    if (iflg) {
        if ((fp = fopen("OPMlist", "w")) == NULL)
            Cleanup("can't open 'OPMlist' for write");
        fprintf(fp, "%d\n", M);
    }
    for (m=0; m<S; m++) {
        scale = Channel[m].Scale;
        GetDsData(&Header, 0, m, scale, In);
        demean(In, T);
        if (Channel[m].ChannelType == TYPE_MEG || Channel[m].ChannelType == TYPE_REF_MAG) {
            for (t=0; t<T; t++)
                Data[m][t] = In[t];
            if (iflg) {
                for (n=0; n<N_OPM; n++)
                    if (!strncmp(OPMname[n], Channel[m].ChannelName, 5)) {
                        fprintf(fp, "%d\n", n);
                        break;
                    }
            }
        } else {        // scale ADC reference channel to volts
            for (t=0; t<T; t++)
                Data[0][t] = ADGAIN * In[t];
        }

		// channel heartbeat
		if(vflg == TRUE) {
			printf(".");
			fflush(stdout);
		}
	}
	if (iflg)
	    fclose(fp);

    // examine reference spectrum -- assume reference is channel 0, for now
    if (vflg) {
        printf(" - done\n");
        printf("searching for reference spectral line up to 100 Hz");
        fflush(stdout);
    }
    for (freq=max=0., f=0; freq<=100.; freq+=df, f++) {
        for (w=0, Ayy=0.; w<NumSegs; w++) {
            for (t=0, tt=w*Step; t<TW; t++, tt++)       // average spectrum over all segments
                Seg[t] = Data[0][tt];
            demean(Seg,TW);
            for (t=0; t<TW; t++)
                y[t] = Hann[t] * Seg[t];
            fftw_execute(RsFwd);
            Yo = Y[f] / bins;
            Ayy += Yo * conj(Yo);
        }
        if (Ayy > max) {
            s = f;
            max = Ayy;
        }

		// frequency bin heartbeat
		if(vflg) {
			printf(".");
			fflush(stdout);
		}
    }
    max /= (double)NumSegs;
    if (vflg) {
        printf(" - done\n");
        printf("maximum reference signal at %4.1f Hz, s=%d, rms=%e\n", (double)s/SegTime, s, sqrt(max));
        fflush(stdout);
    }

    // create list of harmonics up to 100 Hz
    if (vflg) {
        printf("computing ADC & OPM harmonics of reference frequency:\n");
        printf("sensor output (pT or V rms):\n");
        for (h=1; h<=4; h++)
            printf("\t\t%4.1f Hz", (double)(h*s)/SegTime);
        printf("\n");
        fflush(stdout);
    }
    for (m=0; m<S; m++) {
        if (vflg) {
            if (m > 0)
                printf("opm %d:\t", m);
            else
                printf("ADC:\t");
            fflush(stdout);
        }
        for (h=1; h<=4; h++) {
            f = h * s;
            for (w=0, Hxx=0.; w<NumSegs; w++) {
                for (t=0, tt=w*Step; t<TW; t++, tt++)
                    Seg[t] = Data[m][tt];
                demean(Seg, TW);
                for (t=0; t<TW; t++)
                    x[t] = Hann[t] * Seg[t];
                fftw_execute(PsFwd);            // compute channel FFT
                Xo = X[f] / bins;
                Hxx += Xo * conj(Xo);
            }
            Hxx /= (double)NumSegs;
            if (vflg) {
                printf("\t%e", sqrt(Hxx));
			    fflush(stdout);
		    }
        }
        if (vflg) {
            printf("\n");
            fflush(stdout);
        }
    }

    // estimate sensor noise from the minimum power in the range 0-40 Hz
    if (vflg) {
        printf("searching for minimum spectral line up to 40 Hz:\n");
        printf("sensor noise (fT/√Hz):\n");
        fflush(stdout);
    }
    sprintf(OutName, "%s,Noise.txt", SetName);
    if ((fp = fopen(OutName, "w")) == NULL)
        Cleanup("can't open output file for write");
    for (m=0;  m<M; m++) {
        for (freq=0., f=0, min=HUGE; freq<=40.; freq+=df, f++) {
            for (w=0, Axx[m]=0.; w<NumSegs; w++) {
                for (t=0, tt=w*Step; t<TW; t++, tt++)
                    Seg[t] = Data[m+1][tt];     // m+1 skips ADC channel!
                demean(Seg,TW);
                for (t=0; t<TW; t++)
                    x[t] = Hann[t] * Seg[t];
                fftw_execute(PsFwd);
                for (ss=(f-1), m2=0.; ss<=(f+1); ss++) {
                    Xo = X[ss] / bins;          // divide by by number of spectral lines
                    tmp = Xo * conj(Xo);
                    m2 += tmp * tmp;
                }
                Axx[m] += sqrt(m2);
            }
            Axx[m] /= (double)NumSegs;
            if (Axx[m] < min)
                min = Axx[m];

#ifdef EBUG
            // show entire spectrum of each sensor
            printf("%7.3f pT @ %4.1f pT/√Hz\n", 1.0e+12*sqrt(Axx[m]/df), freq);
            fflush(stdout);
#endif
        }

        // save mimimum rms autopower
        fprintf(fp, "%d\t%e\n", m, sqrt(min));      // output rms noise
        fflush(fp);

		// sensor heartbeat
		if(vflg == TRUE) {
            printf("opm %d\t%7.3f\n", m, 1.0e+15*sqrt(min/df));
			fflush(stdout);
		}
    }
    fclose(fp);

    // compute averaged reference autopower & its standard deviation
    if(vflg == TRUE) {
        printf("computing averaged reference autopower");
        fflush(stdout);
    }
    sprintf(OutName, "%s,Ayy.txt", SetName);
    if ((fp = fopen(OutName, "w")) == NULL)
        Cleanup("can't open output file for write");
    for (w=0, Ayy=sx=sx2=0.; w<NumSegs; w++) {      // accumulate statistics over all data segments
    
        // for each reference segment
        for (t=0, tt=w*Step; t<TW; t++, tt++)
            Seg[t] = Data[0][tt];
        demean(Seg, TW);
        for (t=0; t<TW; t++)
            y[t] = Hann[t] * Seg[t];
        fftw_execute(RsFwd);                    // compute reference channel autopower
        for (ss=(s-1), m2=0.; ss<=(s+1); ss++) {
            Yo = Y[ss] / bins;
            SegYo[w] = Yo;                      // save segment spectral line
            tmp = Yo * conj(Yo);
            m2 += tmp * tmp;
        }
        tmp = sqrt(m2);
        sx += tmp;
        sx2 += tmp * tmp;
        Ayy += tmp;
    }
    Ayy /= (double)NumSegs;
    stdev = sqrt((sx2 - (sx * sx / (double)NumSegs)) / (double)(NumSegs - 1)) / (double)NumSegs;
    fprintf(fp, "%e\t%e\n", sqrt(Ayy), stdev);      // output volts rms!
    fclose(fp);

    // compute averaged sensor autopower & its standard deviation
    sprintf(OutName, "%s,AxxStats.txt", SetName);
    if ((fp = fopen(OutName, "w")) == NULL)
        Cleanup("can't open output file for write");

    // for each sensor channel segment
    for (m=0; m<M; m++) {
        for (w=0, Axx[m]=sx=sx2=0.; w<NumSegs; w++) {  // accumulate statistics over all data segments
            for (t=0, tt=w*Step; t<TW; t++, tt++)
                Seg[t] = Data[m+1][tt];         // (skipping reference channel)
            demean(Seg, TW);
            for (t=0; t<TW; t++)
                x[t] = Hann[t] * Seg[t];
            fftw_execute(PsFwd);                // compute sensor autopower
            for (ss=(s-1), m2=0.; ss<=(s+1); ss++) {
                Xo = X[ss] / bins;
                SegXo[m][w] = Xo;
                tmp = Xo * conj(Xo);
                m2 += tmp * tmp;
            }
            tmp = sqrt(m2);
            sx += tmp;
            sx2 += tmp * tmp;
            Axx[m] += tmp;
        }

        // compute sensor autopower & standard deviation
        Axx[m] /= (double)NumSegs;
        stdev = sqrt((sx2 - (sx * sx / (double)NumSegs)) / (double)(NumSegs - 1)) / (double)NumSegs;
        AxxVar[m] = stdev;
        fprintf(fp, "%d\t%e\t%e\n", m, Axx[m], stdev);
    }
    fclose(fp);

    // compute averaged crosspowers & standard deviation
    sprintf(OutName, "%s,Cxy.txt", SetName);
    if ((fp = fopen(OutName, "w")) == NULL)
        Cleanup("can't open output file for write");
    for (m=0; m<M; m++) {
        for (w=0, Cxy[m]=CX0, sx=sx2=0.; w<NumSegs; w++) {  // reminder: Cxy is complex!
            Cxy[m] += SegXo[m][w] * conj(SegYo[w]);
            tmp = cabs(Cxy[m]);
            sx += tmp;
            sx2 += tmp * tmp;
        }

        // compute averaged crosspower & standard deviation
        Cxy[m] /= (double)NumSegs;
        stdev = sqrt((sx2 - (sx * sx / (double)NumSegs)) / (double)(NumSegs - 1)) / (double)NumSegs;
        fprintf(fp, "%d\t%e\t%e\n", m, cabs(Cxy[m]), stdev);
    }
    fclose(fp);

    // compute coherence & gain
    if (vflg) {
        printf(" - done\n");
        printf("computing coherence & gain:\n");
        fflush(stdout);
    }
    sprintf(OutName, "%s,Rxy.txt", SetName);
    if ((fp = fopen(OutName, "w")) == NULL)
        Cleanup("can't open output file for write");
    for (m=0; m<M; m++) {
        Rxy = (Cxy[m] * conj(Cxy[m])) / (Axx[m] * Ayy);                 // coherence (magnitude)  
        Phase = fabs(atan2(cimag(Cxy[m]), creal(Cxy[m])) / M_PI);       // phase angle, normalized by pi
        Gain = Axx[m] / cabs(Cxy[m]);                                   // transfer function (magnitude)
        if (vflg) {
            printf("opm %d:\tAxx=%e\tRxy=%8.6f\tGain=%e\tPhase=%5.1f\n", m+1, Axx[m], Rxy, Gain, 180.*Phase);
            fflush(stdout);
        }
        sign[m] = (180.*Phase < 90.)? 1.: -1.;
        fprintf(fp, "%d\t%8.6f\t%e\t%5.1f\n", m, Rxy, Gain, 180.*Phase);
    }
    fclose(fp);

    // save Axx with its sign (no stats)
    sprintf(OutName, "%s,Axx.txt", SetName);
    if ((fp = fopen(OutName, "w")) == NULL)
        Cleanup("can't open output file for write");
    for (m=0; m<M; m++)
        fprintf(fp, "%d\t%e\n", m, sign[m] * sqrt(Axx[m]));     // output rms!
    fclose(fp);

    // save Axx variance -- use this for noise variance for fitting sensors & dipole coils!
    sprintf(OutName, "%s,AxxVar.txt", SetName);
    if ((fp = fopen(OutName, "w")) == NULL)
        Cleanup("can't open output file for write");
    for (m=0; m<M; m++)
        fprintf(fp, "%d\t%e\n", m, AxxVar[m]);
    fclose(fp);

    // finit!
    if (vflg) {
        printf("\n\nOPM_Stats done\n");
        fflush(stdout);
    }
    exit(0);
}
