// SAMcoreg -- This program coregisters a the MEG dataset sensor coordinate
//  frame to a cortical atlas, derived from segmentation of the cortical surface
//  of structural MRI using Freesurfer. The resulting transform is written into
//  the dataset SAM subdirectory so that subsequent beamformer solutions can read
//  the transform using the parameter keyword "Transform". The name for this file
//  is derived from the cortical atlas name -- e.g., "pial.xfm".
//
// Explanation of SAMcoreg algorithm:
//
//  The algorithm is based upon three principles:
//  (1) The "true" source moment vector is directed approximately perpendicular
//      to the cortical surface. This is termed the "constrained" moment vector.
//  (2) A moment vector for the best signal-to-noise ratio (SNR) can be computed
//      from the MEG data using a generalized eigensystem solution. This is
//      termed the "unconstrained" moment vector. The unconstrained moment vector
//      error is dependent on SNR.
//  (3) The angular error between the constrained and unconstrained moment
//      vectors, over all vertices, will be monotonically related to SNR (z^2).
//      That is, the angular error will be smallest for the largest z^2 values.
//
// This program employs a 6-dimensional conjugate gradient approach. The
//  translations and rotations (Euler angles) are treated as a single vector a[].
//  Rank correlation is computed between the angular error and the LCMV
//  beamformer output signal-to-noise ration, z^2. The angular error is
//  computed from the arc cosine of the absolute value of the dot product
//  of the constrained (vertex normal vector) and the unconstrained moment
//  vector computed from the MEG data at that vertex. The partial derivates
//  - delta correlation vs delta parameter - dr/dX, dr/dY, dr/dZ, dr/dPsi,
//  dr/dTheta, dr/dPhi (drda[6]), where 'r' the rank correlation (either
//  Spearman's rho or Kendall's tau), computed for all vertices meeting the
//  threshold criteria (see below), using MeshScatter. The refined conjugate
//  gradient, is computed uaing the Fletcher-Reeves method, as follows:
//
//      if k=0  p(k)= −g(k), if k=0
//      if k>0  p(k) = −g(k) + β(k)p(k−1)
//
//      where: β(k) = g(k)'g(k) / g(k-1)'g(k-1)
//
//      note that the gradient used in the search is normalized p(k)
//
//      g(k):   conjugate gradient vector (identical to partial derivatives)
//      p(k):   Fletcher-Reeves refined gradient direction
//      k:      pass
//
// The transform vector [a] is incremented by delta and the mesh rank
//  correlation re-evaluated. A new conjugate gradient vector is computed
//  for each improvement in correlation.
//
// The key tunable parameters are:
//  DeltaTrans -- Delta used for computing the translation partial
//      derivatives. This is typically 1 mm.
//  DeltaRotat -- Delta used for computing the rotation partial
//      derivatives. This is typically 2 mrad
//  InitDisp -- Starting displacement -- typically 5 mm/mrad
//  EndDisp -- Ending displacement -- typically 0.01 mm/mrad
//  MinZ -- Cutoff along the z-axis for vertices that are distant
//      from the sensors. This is typically 4 cm (in the CTF
//      coordinate frame.
//  MaxDot -- Maximum dot product of the vertex normal vector with
//      a radial vector to that vertex. Small values of MaxDot
//      eliminate nearly radial normal vectors. This is typically 0.2.
//  MinSpan -- Fraction of vertices sorted by span in ascending order.
//      This is the ratio of the largest to smallest eigenvalue
//      for the generalized eigensystem solution for the unconstrained
//      moment vector. Small spans indicate that the moment
//      vector solution is poorly determined. A MinSpan of 0.5 assures
//      accurate solutions (within the constraint of an underdetermined
//      system.
//
// Additional parameters are available for setting starting & ending
//  displacements along the gradient. These should not be changed.
//  All computations use unscaled SI units. Tesla, meters, radians, etc.
//
// This program assumes that a high-quality atlas of the cortex has been
//  segmented from a volumetric MRI of each subject under study. I strongly
//  recommend using the pial surface minus 3 mm (along the vertex normal),
//  for vertices superior to, say, z=4 cm. The initial surface vertices are
//  generated by FSnormals.py. I also recommend using the 15-30 Hz frequency
//  band where there is good source spatial resolution & adequate S/N. Tuneable
//  parameters are provided in the file coreg.h. The default parameters may be
//  over-ridden by the parameter file.
//
// The outputs of this program "<datasetname>,Transform.txt" -- a running
//  snapshot of the slope versus the translations & rotations, and
//  optionally, for each complete pass, a "Meshnn.txt" file for a scatter
//  plot of z^2 and angular error for each vertex in the atlas.
//
// This version uses the multiple local spheres model forward solution
//
//  flags:
//  -r <run name>           -- MEG Run name
//  -m <parameters>         -- parameter file
//  -p                      -- output passes
//  -v                      -- verbose mode
//
//  Author: Stephen E. Robinson
//      MEG Core Facility
//      NIMH
//

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <math.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <float.h>
#include <string.h>
#include <fcntl.h>
#include <model.h>
#include <samlib.h>
#include <siglib.h>
#include <lfulib.h>
#include <mesh.h>
#include <coreg.h>
#include <DataFiles.h>
#include <SAMfiles.h>
#include <version.h>
#include <samutil.h>
#include <sam_parse.h>
#include <sam_param.h>

/*
#define MINOR_REV   5           // conjugate gradient for maximizing correlation, slope or minimizing chi2
#define MINOR_REV   6           // added optional compile for multisphere -- correlation, onlu
#define MINOR_REV   7           // use SVD instead of LU decomposition because Hessian is rank deficien
#define MINOR_REV   8           // simplified Mesh generation using Atlas2Mesh() & Atlas2RandomMesh()
#define MINOR_REV   9           // creep along gradient by delta
#define MINOR_REV   10          // improved solution for gtadient; reduce delta on failed passes
#define MINOR_REV   11          // added gradient angular change between passes
#define MINOR_REV   12          // added 2nd delta for search along gradient
#define MINOR_REV   13          // refine gradient using the Fletcher-Reeves method for passes > 1
#define MINOR_REV   14          // single delta version
#define MINOR_REV   15          // added separate deltas for partial derivatives translation & rotation
#define MINOR_REV   16          // changed variable names InitTrans & EndTrans to InitDisp & EndDisp, respectively
*/
#define MINOR_REV   17          // changed logic so search continues along gradient for each new gradient computation


int main(
    int         argc,
    char        **argv
) {
    VOXELINFO       *Atlas;             // Atlas[N] - cortical mesh vertices and normals
    MESHINFO        *Mesh;              // Mesh[MAX_VERTEX] -- cortical mesh with solution info
    XFORM           Where;              // translation rotation matrix
    HeaderInfo      Header;             // MEG data header
    PARMINFO        Params;             // analysis parameters
    PARM            *p;                 // pointer into active parameter list
    ChannelInfo     *Channel;           // MEG channel info
    ChannelInfo     *NewChannel;        // new MEG channel info (for tranformation)
    EpochInfo       *Epoch;             // MEG epoch info
    COV_HDR         CovHeader;          // covariance file header
    gsl_matrix      *C;                 // MxM covariance matrix
    gsl_matrix      *Cinv;              // MxM inverse covariance matrix
    gsl_vector      *g1;                // current gradient vector (partial derivatives)
    gsl_vector      *g0;                // previous gradient vector
    gsl_vector      *p1;                // current gradient vector
    gsl_vector      *p0;                // previous gradient vector
    gsl_vector      *gradient;          // conjugate gradient vector
    double          a[6];               // transform origin -- translation & rotation
    double          aNew[6];            // updated parameter vector
    double          Rotation[3];        // rotation (Euler angles)
    double          m0;                 // magnitude of g0
    double          m1;                 // magnitude of g1
    double          beta;               // ratio of g1/g2
    double          *o, Offset[3];      // xyz offsets, for testing convergence         
    double          dt;                 // delta for computing partial derivatives for translation
    double          dr;                 // delta for computing partial derivatives for rotation
    double          delta;              // delta for searching along gradient
    double          x2;                 // chi^2
    double          corr;               // correlation coefficient
    double          slope;              // slope of the regression
    double          rP;                 // correlation for positive delta
    double          rN;                 // correlation for negative delta
    double          rms;                // rms value
    double          Noise;              // noise power
    double          rold;               // old correlation
    double          r;                  // length
    double          dot;                // dot product
    double          angle;              // angle between successive gradients
    int             i;                  // i-index
    int             j;                  // j-index
    int             *ChanIndex;         // ChanIndex[M] -- index of channels used for covariance
    int             M;                  // number of MEG or reference channels used
    int             n;                  // mesh index
    int             Natlas;             // number of atlas vertices read
    int             Nused;              // number of vertices used
    int             v;                  // vector index
    int             pass;               // pass number
    int             ndim = 0;           // pinv dimensions to remove
    int             increase = FALSE;   // flag for at least one successful correlation increase
    int             found = FALSE;      // flag that higher correlation found
    int             rflg = FALSE;       // dataset flag
    int             mflg = FALSE;       // parameter flag
    int             eflg = FALSE;       // command-line error flag
    int             pflg = FALSE;       // output passes
    int             oflg = FALSE;       // optional hidden offset flag
    int             vflg = FALSE;       // verbose mode flag (default: silent)
    time_t          secs;               // random number seed
    char            *name;              // random filename
    char            *DSName = NULL;     // dataset name
    char            AtlasPath[256];     // atlas path name
    char            CovPath[256];       // covariance file path name
    char            SAMpath[256];       // data set path
    char            fpath[256];         // general path name
    char            fname[64];          // general file name
    char            *CovName = NULL;    // name used in covariance files
    char            *XformName = NULL;  // name used for Xform outputs
    char            Prefix[64];         // dataset prefix code
    FILE            *fp;                // log file pointer
    FILE            *sfp;               // scattergram file pointer

    // Register the parameters used by this program.
    reg_usage(argv[0], MINOR_REV, __DATE__);
    reg_std_parm();

    reg_parm("CovName");
    reg_parm("MRIDirectory");
    reg_parm("MRIPattern");
    reg_parm("PrefixLength");
    reg_parm("AtlasName");
    reg_parm("XformName");
    reg_parm("CovBand");
    reg_parm("Pinv");
    reg_parm("NumVerts");
    reg_parm("MinZ");
    reg_parm("MinSpan");
    reg_parm("MaxAngle");
    reg_parm("MaxDot");
    reg_parm("MinSNR");
    reg_parm("ShowPasses");
    reg_parm("DeltaTranslate");
    reg_parm("DeltaRotate");
    reg_parm("InitDisplacement");
    reg_parm("EndDisplacement");
    reg_parm("Offset");

    // Parse the arguments, environment variables, and parameter files.
    do_parse_args(argc, argv);

    // Set default parameters, where different from the global defaults.
    new_params(&Params);

    // Extract specified parameters.
    eflg = do_get_params(&Params);

    p = get_parm("DataSet");
    if (p->set) {
        rflg = TRUE;
        DSName = copy_string(Params.DataSetName);   // DSName is modified
    }
    p = get_parm("verbose");
    vflg = p->set;
    p = get_parm("param");
    mflg = p->set;
    p = get_parm("Pinv");
    if (p->set) {
        ndim = *(int *)p->ptr;
    }
    p = get_parm("ShowPasses");
    pflg = p->set;

    p = get_parm("Offset");     // optional hidden translation offset (mm) for testing convergence
    if (p->set) {
        o = (double *)p->ptr;
        Offset[aX] = o[0];
        Offset[aY] = o[1];
        Offset[aZ] = o[2];
        for (v=aX; v<=aZ; v++)  // convert offet mm to m
            Offset[v] *= 0.001;
        oflg = TRUE;
    }

    if (!rflg || !mflg || eflg) {
        msg("dataset (-r) and parameter file (-m) are required\n");
        do_help();  // doesn't return
    }

    CovName = Params.ParmName;
    p = get_parm("CovName");
    if (p->set) {
        CovName = (char *)p->ptr;
    }

    XformName = "";
    p = get_parm("XformName");
    if (p->set) {
        XformName = (char *)p->ptr;
    }

    // set the random number seed
    time(&secs);
    srandom((unsigned int)secs);

    // announce
    if (vflg) {
        printf("S A M   C O R E G I S T R A T I O N\t%s\n", PRG_REV);
        printf("*** Coregistration by maximizing correlation ***\n");
        printf("opening data file");
        fflush(stdout);
    }

    // get data with sensor structures
    name = copy_string(DSName);
    GetDsInfo(DSName, &Header, &Channel, &Epoch, NULL, TRUE);
    GetDsInfo(name, &Header, &NewChannel, &Epoch, NULL, TRUE);

    // optionally translate the sensor array by 'Offset'
    if (oflg) {         // modify 'Channel' positions if offsets are introduced
        Where.Translate[X_] = Offset[X_];
        Where.Translate[Y_] = Offset[Y_];
        Where.Translate[Z_] = Offset[Z_];
        Rotation[PSI_] = 0.;
        Rotation[THETA_] = 0.;
        Rotation[PHI_] = 0.;
        EtoR(Rotation, Where.Rotate);
        MoveMEG(NewChannel, Channel, Header.SqIndex, Header.NumSquid, &Where);
    }
    GetFilePath(DSDIR, SAMpath, sizeof(SAMpath), &Params, "SAM", TRUE);
    free(name);

    if (vflg) {
        printf(" - done\n");
        fflush(stdout);
    }

    // set dataset constants
    M = Header.NumPri;

    p = get_parm("CovBand");
    if (!p->set) {
        fatalerr("CovBand req'd");
    }
    p = get_parm("MRIDirectory");
    if (!p->set) {
        fatalerr("MRIDirectory req'd");
    }
    p = get_parm("AtlasName");
    if (!p->set) {
        fatalerr("AtlasName req'd");
    }

    // summarize coregistration parameters
    if (vflg) {
        printf("\tMultiple Local Spheres\n");
        if (Params.MaxVertex > 0)
            printf("\tNumber of Vertices=%d\n", Params.MaxVertex);
        else
            printf("\tUsing all Vertices in Mesh\n");
        printf("\tTranslation Delta=%5.3f mm\n", 1000.*Params.DeltaTrans);
        printf("\tRotation Delta=%5.3f mrad\n", 1000.*Params.DeltaRotat);
        printf("\tInitial Displacement=%5.3f mm/mrad\n", 1000.*Params.InitDisp);
        printf("\tEnding Displacement=%5.3f mm/mrad\n", 1000.*Params.EndDisp);
        if (oflg)
            printf("\tOffsets: %6.3f %6.3f %6.3f mm\n", 1000.*Offset[X_], 1000.*Offset[Y_], 1000.*Offset[Z_]);
        printf("\tMinimum z=%4.1f cm\n", 100.*Params.MinZ);
        printf("\tMinimum Span=%5.3f\n", Params.MinSpan);
        printf("\tMinimum SNR=%5.3f x maximum\n", Params.MinSNR);
        printf("\tMaximum dot product=%5.3f\n", Params.MaxDot);
        fflush(stdout);
    }

    // allocate gsl vectors for gradients
    g0 = gsl_vector_alloc(6);
    g1 = gsl_vector_alloc(6);
    p0 = gsl_vector_alloc(6);
    p1 = gsl_vector_alloc(6);
    gradient = gsl_vector_alloc(6);

    // set prefix to NumPrefix characters (default is 8-character dataset hashcode)
    memset(Prefix, 0, sizeof(Prefix));
    memcpy(Prefix, Header.SetName, Params.NumPrefix);
    if (vflg) {
        printf("setting multisphere model");
        fflush(stdout);
    }
    SetSphereOrigins(DSName, &Header, Channel, MSPHERE);

    // get atlas
    if (vflg) {
        printf(" - done\n");
        printf("reading atlas file");
        fflush(stdout);
    }
    p = get_parm("AtlasName");
    name = (char *)p->ptr;
    GetMRIPath(AtlasPath, sizeof(AtlasPath), &Params, name, TRUE);
    GetAtlas(AtlasPath, &Atlas, &Natlas);
    if (vflg) {
        printf(" - read %d vertices\n", Natlas);
        fflush(stdout);
    }

    // allocate Mesh & Result structure
    if(Params.MaxVertex > 0) {          // select MaxVertex random mesh elements from N vertices
        Mesh = new_arrayE(MESHINFO, Params.MaxVertex, "Mesh[]");
        Atlas2RandomMesh(Atlas, Natlas, Mesh, &Params);
    } else {                            // select all vertices
        Mesh = new_arrayE(MESHINFO, Natlas, "Mesh[]");
        Atlas2Mesh(Atlas, Natlas, Mesh, &Params);
    }
    free(Atlas);
    if (vflg) {
        printf("selected %d Mesh vertices\n", Params.MaxVertex);
        fflush(stdout);
    }

    // read covariance file & mean sensor noise
    if (vflg) {
        printf("reading covariance file & mean sensor noise");
        fflush(stdout);
    }
    snprintf(fpath, sizeof(fpath), "%s,%-d-%-dHz", CovName, (int)Params.CovHP, (int)Params.CovLP);
    GetFilePath("%d/SAM/%s/Global.cov", CovPath, sizeof(CovPath), &Params, fpath, 0);
    C = gsl_matrix_alloc(M, M);
    Cinv = gsl_matrix_alloc(M, M);
    GetCov(CovPath, &CovHeader, &ChanIndex, C);
    if (ndim > 0) {
        pinvN(C, Cinv, ndim);
    } else {
        pinv(C, Cinv);
    }

    name = copy_string(fpath);
    GetFilePath("%d/SAM/%s/Global_Noise", fpath, sizeof(fpath), &Params, name, 0);
    free(name);
    if ((fp = fopen(fpath, "r")) == NULL)
        Cleanup("can't open 'Global_Noise'");
    if (fscanf(fp, "%lf", &Noise) != 1)
        Cleanup("can't read 'Global_Noise'");
    fclose(fp);

    // open log file, & write header
    snprintf(fpath, sizeof(fpath), "%s,%sTransform.txt", Header.SetName,XformName);
    if ((fp = fopen(fpath, "w")) == NULL)
        Cleanup("can't open transform summary for write");
    fprintf(fp, "Dataset '%s'\n", Header.SetName);
    fprintf(fp, "Number of SQUID Sensors: %d (%d Primary Sensors)\n", Header.NumSquid, M);
    fprintf(fp, "Coregistration in %6.3f to %6.3f Hz bandpass\n", Params.CovHP, Params.CovLP);
    fprintf(fp, "Cortical Mesh Coregistration Parameters:\n");
    fprintf(fp, "\tMultiple Local Spheres\n");
    fprintf(fp, "\tUsing %d Mesh Vertices\n", Params.MaxVertex);
    fprintf(fp, "\tTranslation Delta=%5.3f mm\n", 1000.*Params.DeltaTrans);
    fprintf(fp, "\tRotation Delta=%5.3f mrad\n", 1000.*Params.DeltaRotat);
    fprintf(fp, "\tInitial Displacement=%5.3f mm/mrad\n", 1000.*Params.InitDisp);
    fprintf(fp, "\tEnding Displacement=%5.3f mm/mrad\n", 1000.*Params.EndDisp);
    if (oflg)
        fprintf(fp, "\tOffsets: %6.3f %6.3f %6.3f mm\n", 1000.*Offset[X_], 1000.*Offset[Y_], 1000.*Offset[Z_]);
    fprintf(fp, "\tMinimum Span=%5.3f\n", Params.MinSpan);
    fprintf(fp, "\tMinimum z=%4.1f cm\n", 100.*Params.MinZ);
    fprintf(fp, "\tMinimum SNR=%5.3f x maximum\n", Params.MinSNR);
    fprintf(fp, "\tMaximum dot product=%5.3f\n", Params.MaxDot);
    fflush(fp);

    // initializations: Note that we can introduce initial conditions for testing convergence
    for (v=0; v<6; v++)
        a[v] = 0.;
    pass = 0;

    // compute initial MeshScatter() correlation, with no displacement (* optional offset)
    if (vflg) {
        printf(" - done\n");
        printf("evaluating initial mesh correlation");
        fflush(stdout);
    }
    MeshScatterMS(&Header, Channel, C, Cinv, Noise, Mesh, &Params, &Nused, &slope, &x2, &corr, &rms);
    rold = corr;
    if (vflg) {
        printf(" - done\n");
        printf("\tcorr=%f, slope=%f, chi^2=%f rms=%f, n=%d\n", corr, slope, x2, rms, Nused);
        fflush(stdout);
    }

    // output transform header
    fprintf(fp, "\n                         Transform Evaluations\n");
    fprintf(fp, "      X       Y       Z     Psi   Theta     Phi     corr    slope    chi^2      rms      n   pass\n");
    fflush(fp);

    // log initial translation
    //               x     y     z   psi theta   phi  corr slope   x^2   rms   n pass
    fprintf(fp, "%7.3f %7.3f %7.3f %7.3f %7.3f %7.3f %8.6f %8.6f %8.6f %8.6f %6d %6d\n",
      1000.*a[aX], 1000.*a[aY], 1000.*a[aZ], 1000.*a[aPSI], 1000.*a[aTHETA], 1000.*a[aPHI],
      corr, slope, x2, rms, Nused, pass);
    fflush(fp);

    // save scatter plot
    if (pflg) {
        sprintf(fpath, "%s%s%02d.txt", Prefix, XformName, pass);
        if ((sfp = fopen(fpath, "w")) == NULL)
            Cleanup("can't open scatter plot file for write");
        for (n=0; n<Params.MaxVertex; n++)      // save angular error (radians) vs z^2
            if(Mesh[n].angle >= 0.)
                fprintf(sfp, "%5.3f\t%6.3f\n", Mesh[n].angle, Mesh[n].z2);
        fclose(sfp);
    }

    // begin conjugate gradient search maximizing correlation
    dt = Params.DeltaTrans;
    dr = Params.DeltaRotat;
    pass = 1;
    do {

        // evaluate partial derivatives symmetrically around current transform origin a[]
        if (vflg) {
            printf("evaluating partial derivatives:\n");
            fflush(stdout);
        }
        for (i=0; i<6; i++) {

            // evaluate correlation for positive delta
            for (j=0; j<6; j++)
                if (j == i) {
                    if (j < 3)
                        aNew[j] = a[j] + dt;
                    else
                        aNew[j] = a[j] + dr;
                } else {
                    aNew[j] = a[j];
                }
            Where.Translate[X_] = aNew[aX];
            Where.Translate[Y_] = aNew[aY];
            Where.Translate[Z_] = aNew[aZ];
            Rotation[PSI_] = aNew[aPSI];
            Rotation[THETA_] = aNew[aTHETA];
            Rotation[PHI_] = aNew[aPHI];
            EtoR(Rotation, Where.Rotate);
#ifdef EBUG
            printf("displacements (+): %f %f %f %f %f %f\n", 1000.*aNew[aX], 1000.*aNew[aY], 1000.*aNew[aZ], 1000.*aNew[aPSI], 1000.*aNew[aTHETA], 1000.*aNew[aPHI]);
            fflush(stdout);
#endif
            MoveMEG(Channel, NewChannel, Header.SqIndex, Header.NumSquid, &Where);
            MeshScatterMS(&Header, NewChannel, C, Cinv, Noise, Mesh, &Params, &Nused, &slope, &x2, &rP, &rms);

            // evaluate mesh correlation for negative delta
            for (j=0; j<6; j++)
                if (j == i) {
                    if (j < 3)
                        aNew[j] = a[j] - dt;
                    else
                        aNew[j] = a[j] - dr;
                } else {
                    aNew[j] = a[j];
                }
            Where.Translate[X_] = aNew[aX];
            Where.Translate[Y_] = aNew[aY];
            Where.Translate[Z_] = aNew[aZ];
            Rotation[PSI_] = aNew[aPSI];
            Rotation[THETA_] = aNew[aTHETA];
            Rotation[PHI_] = aNew[aPHI];
            EtoR(Rotation, Where.Rotate);
#ifdef EBUG
            printf("displacements (-): %f %f %f %f %f %f\n", 1000.*aNew[aX], 1000.*aNew[aY], 1000.*aNew[aZ], 1000.*aNew[aPSI], 1000.*aNew[aTHETA], 1000.*aNew[aPHI]);
            fflus(stdout);
#endif
            MoveMEG(Channel, NewChannel, Header.SqIndex, Header.NumSquid, &Where);
            MeshScatterMS(&Header, NewChannel, C, Cinv, Noise, Mesh, &Params, &Nused, &slope, &x2, &rN, &rms);

            // finally, compute partial derivatives
            if (i < 3)
                gsl_vector_set(g1, i, (atanh(rP) - atanh(rN)) / (2. * dt));
            else
                gsl_vector_set(g1, i, (atanh(rP) - atanh(rN)) / (2. * dr));

            // show gradient progress
            if (vflg) {
                switch (i) {
                    case 0: printf("\tdr/dx=%10.6f\n", gsl_vector_get(g1, 0)); break;
                    case 1: printf("\tdr/dy=%10.6f\n", gsl_vector_get(g1, 1)); break;
                    case 2: printf("\tdr/dz=%10.6f\n", gsl_vector_get(g1, 2)); break;
                    case 3: printf("\tdr/d\u03c8=%10.6f\n", gsl_vector_get(g1, 3)); break;    // unicode psi
                    case 4: printf("\tdr/d\u03b8=%10.6f\n", gsl_vector_get(g1, 4)); break;    // unicode theta
                    case 5: printf("\tdr/d\u03c6=%10.6f\n", gsl_vector_get(g1, 5)); break;    // unicode phi
                    default: break;
                }
                fflush(stdout);
            }
        }

        // compute conjugate gradient from drda
        if (vflg) {
            printf("compute conjugate gradient vector\n");
            fflush(stdout);
        }   
        if (pass == 1) {                        // 1st pass, p1 is g1 (partial derivatives)
            gsl_blas_dcopy(g1, p1);             // p1 = g1

            // copy this pass to previous pass values
            gsl_blas_dcopy(g1, g0);             // g0 = g1
            gsl_blas_dcopy(p1, p0);             // p0 = p1
            angle = 0.;                         // by definition

        } else {                                // subsequent passes, p1 is g1 - beta * p0
            // compute beta
            gsl_blas_ddot(g1, g1, &m1);         // m1 = g1'g1
            gsl_blas_ddot(g0, g0, &m0);         // m0 = g0'g0
            beta = m1 / m0;                     // beta = m1 / m0

            // compute Fletcher-Reeves direction
            gsl_blas_dcopy(g1, p1);             // p1 = g1
            gsl_blas_daxpy(-beta, p0, p1);      // p1 = g1 - beta * p0

            // compare angle between p1 & p0
            gsl_blas_ddot(p1, p0, &dot);
            r = 1. / sqrt(fabs(dot));
            if (r > 1.)
                r = 1.;
            angle = 180. * acos(r) / M_PI;

            // copy this pass to previous pass values
            gsl_blas_dcopy(g1, g0);             // g0 = g1
            gsl_blas_dcopy(p1, p0);             // p0 = p1
        }

        // normalize conjugate gradient
        m1 = gsl_blas_dnrm2(p1);                // m1 = ||p1||^-2
        gsl_blas_dcopy(p1, gradient);
        gsl_blas_dscal(1./m1, gradient);
        if (vflg) {
            printf("gradient:");
            for(v=0; v<6; v++)
                printf(" %9.6f", gsl_vector_get(gradient, v));
            printf("\nmagnitude before normalization: %10.6f\n", m1);
            printf("angular change: %7.3f degrees\n", angle);
            fflush(stdout);
        }

        // evaluate transform & step along conjugate gradient until rank correlation drops
        increase = FALSE;       // set flag to detect if there was an increase for this gradient
        delta = Params.InitDisp;
        do {

            // set found false for each try along gradient
            found = FALSE;

            // announce
            if (vflg) {
                printf("evaluating new transform for delta %5.3f mm/mrad:\n", 1000.*delta);
                fflush(stdout);
            }

            // move in direction of conjugate gradient
            for (v=0; v<6; v++)
                aNew[v] = a[v] + delta * gsl_vector_get(gradient, v);
            Where.Translate[X_] = aNew[aX];
            Where.Translate[Y_] = aNew[aY];
            Where.Translate[Z_] = aNew[aZ];
            Rotation[PSI_] = aNew[aPSI];
            Rotation[THETA_] = aNew[aTHETA];
            Rotation[PHI_] = aNew[aPHI];
            EtoR(Rotation, Where.Rotate);
            MoveMEG(Channel, NewChannel, Header.SqIndex, Header.NumSquid, &Where);
            MeshScatterMS(&Header, NewChannel, C, Cinv, Noise, Mesh, &Params, &Nused, &slope, &x2, &corr, &rms);
            if (vflg) {
                printf("\tcorr=%f, slope=%f, chi^2=%f rms=%f, n=%d\n", corr, slope, x2, rms, Nused);
                fflush(stdout);
            }
            
            // if successful, log pass & update a[]
            if (corr > rold) {

                // log the test transform
                fprintf(fp, "%7.3f %7.3f %7.3f %7.3f %7.3f %7.3f %8.6f %8.6f %8.6f %8.6f %6d %6d\n",
                  1000.*aNew[aX], 1000.*aNew[aY], 1000.*aNew[aZ], 1000.*aNew[aPSI], 1000.*aNew[aTHETA], 1000.*aNew[aPHI],
                  corr, slope, x2, rms, Nused, pass);
                fflush(fp);

                // if req'd, save pass delta angle vs. z^2 scattergram for future analysis
                if (pflg) {
                    snprintf(fpath, sizeof(fpath), "%s%s%02d.txt", Prefix, XformName, pass);
                    if ((sfp = fopen(fpath, "w")) == NULL)
                        Cleanup("can't open scatter plot file for write");
                    for (n=0; n<Params.MaxVertex; n++)          // save angular error (radians) vs z^2
                        if(Mesh[n].angle >= 0.)
                            fprintf(sfp, "%5.3f\t%6.3f\n", Mesh[n].angle, Mesh[n].z2);
                    fclose(sfp);
                }

                // update a[]
                for (v=0; v<6; v++)
                    a[v] = aNew[v];
                rold = corr;
                pass++;
                increase = TRUE;    // signify that we've found at least one increase per new gradient
                found = TRUE;
            } else {                // failed to improve correlation
                if (!found)         // decrease delta if no improvement in correlation
                    delta /= 2.;
            }
                
        } while (delta > Params.EndDisp);
#ifdef EBUG
        printf("found=%d, increase=%d, delta=%f\n", found, increase, 1000.*delta);
        fflush(stdout);
#endif
    } while (increase);             // compute a new gradient until no increase for all deltas

    // replicate using best a[] vector for logging current values
    Where.Translate[X_] = a[aX];
    Where.Translate[Y_] = a[aY];
    Where.Translate[Z_] = a[aZ];
    Rotation[PSI_] = a[aPSI];
    Rotation[THETA_] = a[aTHETA];
    Rotation[PHI_] = a[aPHI];
    EtoR(Rotation, Where.Rotate);
    MoveMEG(Channel, NewChannel, Header.SqIndex, Header.NumSquid, &Where);
    MeshScatterMS(&Header, NewChannel, C, Cinv, Noise, Mesh, &Params, &Nused, &slope, &x2, &corr, &rms);

    // log the final transform
    fprintf(fp, "%7.3f %7.3f %7.3f %7.3f %7.3f %7.3f %8.6f %8.6f %8.6f %8.6f %6d\n",
      1000.*a[aX], 1000.*a[aY], 1000.*a[aZ], 1000.*a[aPSI], 1000.*a[aTHETA], 1000.*a[aPHI],
      corr, slope, x2, rms, Nused);
    fflush(fp);

    // if req'd, save final scattergram for future analysis
    if (pflg) {
        snprintf(fpath, sizeof(fpath), "%s%s%02d.txt", Prefix, XformName, pass);
        if ((sfp = fopen(fpath, "w")) == NULL)
            Cleanup("can't open scatter plot file for write");
        for (n=0; n<Params.MaxVertex; n++)          // save angular error (radians) vs z^2
            if(Mesh[n].angle >= 0.)
                fprintf(sfp, "%5.3f\t%6.3f\n", Mesh[n].angle, Mesh[n].z2);
        fclose(sfp);
    }

    // report final transform
    fprintf(fp, "\nBest Metric=%8.6f at:\n", corr);
    fprintf(fp, "\t%7.3f, %7.3f, %7.3f mm, & %7.3f, %7.3f, %7.3f milliradians\n",
      1000.*a[aX], 1000.*a[aY], 1000.*a[aZ], 1000.*a[aPSI], 1000.*a[aTHETA], 1000.*a[aPHI]);
    fclose(fp);

    // save transform parameters (unscaled) in SAM subdirectory
    sprintf(fname, "%s%s", XformName, Params.AtlasName);
    GetFilePath("%d/SAM/%s.xfm", fpath, sizeof(fpath), &Params, fname, 0);
    fp = fileopen(fpath, "w");
    fprintf(fp, "%e\t%e\t%e\t%e\t%e\t%e", a[aX], a[aY], a[aZ], a[aPSI], a[aTHETA], a[aPHI]);
    fclose(fp);

    // finit
    log_params(SAMpath);;
    if (vflg) {
        printf(" - done\n");
        printf("'%s' done\n", Progname);
    }
    exit(0);
}
