// SAMcoreg -- Coregistration of a MEG dataset with a cortical atlas
//  derived from Freesurfer. This is a brute-force algorithm that
//  iterates on successive translations & rotations of the MEG sensors
//  (both primary & reference) to maximize the [inverse of the] product-
//  moment correlation coefficent of angular error with z^2. The
//  resulting transformations are written into the SAM subdirectory
//  of the dataset so that subsequent beamformer solutions can read the
//  transformation using the parameter keyword "Transform". The default
//  name for this file is "Transform.xfm." The cortical normals are read
//  as the _constrained_ moment vector, while the linear solution for
//  the moment vector from the MEG data is the _unconstrained_ moment
//  vector. This version does independent translation & rotation searches.
//  Uncoupling translation & rotation reduces their interaction.
//
// Explanation of SAMcoreg algorithm:
//  The algorithm is based upon the relationship between LCMV beamformer
//  moment vector error and z^2. The monotonic [statistical] relationship
//  between source resolution & z^2 has been published. The accuracy of the
//  linear solution for the moment vector is a function of z^2.
//
// This program assumes that a high-quality atlas of the cortex has been
//  segmented from an MRI of each subject being studied. I strongly recommend
//  using the pial surface, for vertices superior to z=5 cm. The initial
//  surface vertices are generated by FSnormals.py. I also recommend using
//  the 4-40 Hz frequency band where there is good source spatial resolution
//  & adequate S/N. Tuneable parameters are provided in the file coreg.h.
//  The default parameters may be over-ridden by the .param file.
//
// In this version, FreeSurfer vertices are read using 'GetAtlas()'. A subset
//  of these vertices are selected by 'RandomMesh()' that selects MAX_VERTEX
//  vertices. See coreg.h.
// The verticies are additionally constrained to be superior to 'MIN_Z' with
//  a maximum dot product of MAX_DOT, to select sulcal vertices (see coreg.h).
//  For each pass, 'RandomMesh()' is invoked to select a _new_ set of vertices.
//  Multiple passes will anneal the results such that it is as if the entire
//  pial atlas was coregistered.
//
// For each vertex in the atlas, two LCMV beamformer solutions are performed
//  - one using only the MEG data to solve the linear solution for the
//  unconstrained moment vector and the other using the normal vector for that
//  vertex to compute z^2. The difference between these two vectors is the
//  "angular error". The variance of this error over the vertices) should have
//  a monotonic relationship with z^2 if the MEG and MRI are coregistered. In
//  this version, the negative product-moment correlation of angular error
//  (the arc cosine of the dot product of the unconstrained & constrained
//  moment vectors) with z^2 from the constrained solution is used.
//
// For each pass, the program selects the transform giving the maximum
//  correlation. Since a different random mesh is selected for each pass, one
//  cannot expect correlation to improve with each pass. However, the average
//  values of translation and rotation will converge to their best values.
//
// The outputs of this program "<datasetname>,Transform.txt" -- a running
//  snapshot of the correlation versus the translations & rotations, and
//  optionally, for each complete pass, a "Meshnn.txt" file for a scatter
//  plot of z^2 and angular error for each vertex in the atlas.
//
// This version uses the realistic head model (Nolte) forward solution!
//
// SAMcoreg is slow but may be accelerated in future releases, using CUDA.
//
//
//  flags:
//  -r <run name>           -- MEG Run name
//  -m <parameters>         -- parameter file
//  -p                      -- output passes
//  -v                      -- verbose mode
//
//  Author: Stephen E. Robinson
//      MEG Core Facility
//      NIMH
//

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <math.h>
#include <float.h>
#include <string.h>
#include <fcntl.h>
#include <model.h>
#include <samlib.h>
#include <siglib.h>
#include <lfulib.h>
#include <mesh.h>
#include <coreg.h>
#include <DataFiles.h>
#include <SAMfiles.h>
#include <version.h>
#include "samutil.h"
#include "sam_parse.h"

#define MINOR_REV   20                  // old REV 2
#define PASS_LIMIT  100                 // number of passes for a given step size
#define DOTRANS TRUE

typedef struct {
    double  To[3];
    double  Ro[3];
} PASS_RESULTS;

#define aX          0
#define aY          1
#define aZ          2
#define aPSI        3 
#define aTHETA      4 
#define aPHI        5 

int main(
    int         argc,
    char        **argv
) {
    VOXELINFO       *Atlas;             // Atlas[N] - cortical mesh vertices and normals
    MESHINFO        *Mesh;              // Mesh[MAX_VERTEX] -- cortical mesh with solution info
    XFORM           Where;              // translation rotation matrix
    HeaderInfo      Header;             // MEG data header
    HULL            Hull;               // cortical hull model -- vertices & normals
    PARMINFO        Params;             // analysis parameters
    PARM            *p;                 // pointer into active parameter list
    ChannelInfo     *Channel;           // MEG channel info
    ChannelInfo     *NewChannel;        // new MEG channel info (for tranformation)
    EpochInfo       *Epoch;             // MEG epoch info
    COV_HDR         CovHeader;          // covariance file header
    PASS_RESULTS    *Best;              // Best[PASS_LIMIT] -- structure for pass results
    GIIATLAS        *atlas;             // gifti atlas info
    gsl_matrix      *C;                 // MxM covariance matrix
    gsl_matrix      *Cinv;              // MxM inverse covariance matrix
    double          MinSpan;               // Max/Min eigenvalue ratio
    double          Translation[3];     // displacement (translation from origin)
    double          Rotation[3];        // rotation (Euler angles)
    double          *o, Offset[6];      // xyz offsets, for testing
    double          Vo[3];              // Atlas origin
    double          x2;                 // chi^2
    double          Tmax[3];            // translation at maximum correlation
    double          Rmax[3];            // rotation at maximum correlation
    double          Tdisp;              // current translation displacement
    double          Rdisp;              // current rotation displacement
    double          correlation;        // correlation coefficient of z^2 vs. angular error
    double          slope;              // slope of the regression
    double          rms;                // rms value
    double          product;            // product of rms * corr
    double          Noise;              // noise power
    double          zbest;              // best current rms S/N
    double          zprior;             // prior rms S/N
    double          rbest;              // best current correlation
    double          sbest;              // best current slope
    double          xbest;              // best current chi-sq
    double          pbest;              // best current rms * correlation
    double          rprior;             // prior correlation
    double          sprior;             // prior slope
    double          r2;                 // r^2
    double          vmin[3];            // minimum surface vertex coordinates
    double          vmax[3];            // maximum surface vertex coordinates
    double          To[3];              // translation origin -- starting coordinate
    double          Ro[3];              // rotation origin -- starting Euler angles
    double          PriorTo[3];         // prior average translation origin
    double          PriorRo[3];         // prior average rotation origin
    //double          Tstep;              // translation step-size (m)
    //double          Rstep;              // rotation step-size (rad)
    double          tmp;                // useful variable!
    //double        total_error;          // ACN - also look at total angular error
    //double        mean_error;
    int             i;
    int             passcounter;        // ACN - counter for passes
    int             *ChanIndex;         // ChanIndex[M] -- index of channels used for covariance
    int             M;                  // number of MEG or reference channels used
    int             Natlas;             // number of mesh elements
    int             S;                  // number of physical SQUID channels
    int             n;                  // mesh index
    int             Nused;              // number of voxels used
    int             v;                  // vector index
    int             stp;                // step pass number
    int             nstop;              // stop count for reducing deltas
    int             ndim = 0;           // pinv dimensions to remove
    int             rflg = FALSE;       // dataset flag
    int             mflg = FALSE;       // parameter flag
    int             gflg = FALSE;       // gifti format atlas flag
    int             eflg = FALSE;       // command-line error flag
    int             pflg = FALSE;       // output passes
    int             oflg = FALSE;       // optional hidden offset flag
    int             vflg = FALSE;       // verbose mode flag (default: silent)
    time_t          secs;               // random number seed
    char            *name;              // random filename
    char            *DSName = NULL;     // dataset name
    char            AtlasPath[256];     // atlas path name
    char            CovPath[256];       // covariance file path name
    char            HullPath[256];      // hull path name
    char            SAMpath[256];       // data set path
    char            fpath[256];         // general path name
    char            fname[256];         // general path name
    char            *CovName = NULL;    // name used in covariance files
    char            *XformName = NULL;  // name used for Xform outputs
    char            Prefix[64];         // dataset prefix code
    FILE            *fp;                // log file pointer
    FILE            *sfp;               // scattergram file pointer

    // Allison parameters //

//  double      Trans_max[] = { 0.008, 0.003, 0.0015, 0.0006, 0.0003 };
//  double      Trans_step[] = { 0.002, 0.001, 0.0005, 0.0002, 0.0001 };
//  double      Rot_max[] = { 0.008, 0.003, 0.0015, 0.0006, 0.0003 };
//  double      Rot_step[] = { 0.002, 0.001, 0.0005, 0.0002, 0.0001 };
//    double        Trans_max[] = { 0.010, 0.005, 0.002, 0.001, 0.0005 };
//    double        Trans_step[] = { 0.005, 0.0025, 0.001, 0.0005, 0.00025 };
//    double        Rot_max[] = { 0.1, 0.05, 0.02, 0.01, 0.005 };
//    double        Rot_step[] = { 0.05, 0.025, 0.01, 0.005, 0.0025 };
    double      Trans_max[] = { 0.015 };
    double      Trans_step[] = { 0.0025 };
//    double      Rot_max[] = { 0.04 };
//    double      Rot_step[] = { 0.005 };
    double      Rot_max[] = { 0.00 };
    double      Rot_step[] = { 0.00 };
//    double      Trans_max[] = { 0.0 };
////    double      Trans_step[] = { 0.0 };


    // Register the parameters used by this program.

    reg_usage(argv[0], MINOR_REV, __DATE__);
    reg_std_parm();

    reg_parm("CovName");
    reg_parm("MRIDirectory");
    reg_parm("MRIPattern");
    reg_parm("PrefixLength");
    reg_parm("HullName");
    reg_parm("AtlasName");
    reg_parm("XformName");
    reg_parm("CovBand");
    reg_parm("Pinv");

    reg_parm("Model");
    reg_parm("Order");
//    reg_parm("doMultiSphere");  // since Model isn't allowed on the command line
//    reg_parm("doNolte");        // allow --domulti or --donolte

    reg_parm("NumVerts");
    reg_parm("MinSpan");
    reg_parm("MinZ");
    reg_parm("MaxAngle");
    reg_parm("MaxDot");
    reg_parm("MinSNR");
    reg_parm("ShowPasses");
    //reg_parm("InitTranslate");
    //reg_parm("InitRotate");
    //reg_parm("EndTranslate");
    //reg_parm("EndRotate");
    reg_parm("OffsetTrans");
    reg_parm("OffsetRot");

    // Parse the arguments, environment variables, and parameter files.

    do_parse_args(argc, argv);

    // Set default parameters, where different from the global defaults.

    new_params(&Params);
    Params.Order = DEFAULT_ORDER;

    // Extract specified parameters.

    eflg = do_get_params(&Params);

    p = get_parm("DataSet");
    if (p->set) {
        rflg = TRUE;
        DSName = copy_string(Params.DataSetName);   // DSName is modified
    }
    p = get_parm("verbose");
    vflg = p->set;
    p = get_parm("param");
    mflg = p->set;
    p = get_parm("Pinv");
    if (p->set) {
        ndim = *(int *)p->ptr;
    }
    p = get_parm("MinSpan");
    if (p->set) {
        MinSpan = *(double *)p->ptr;
        Params.MinSpan = MinSpan;
    }
    p = get_parm("ShowPasses");
    pflg = p->set;

    p = get_parm("OffsetTrans");     // optional hidden translation offset (mm) for testing convergence
    if (p->set) {
        o = (double *)p->ptr;
        Offset[aX] = o[0];
        Offset[aY] = o[1];
        Offset[aZ] = o[2];
        Offset[aPSI] = 0.;
        Offset[aTHETA] = 0.;
        Offset[aPHI] = 0.;
        for (v=aX; v<=aZ; v++)  // convert offet mm/mrad to m/rad
            Offset[v] *= 0.001;
        oflg = TRUE;
    }
    p = get_parm("OffsetRot");     // optional hidden translation offset (mm) for testing convergence
    if (p->set) {
        o = (double *)p->ptr;
        Offset[aPSI] = o[0];
        Offset[aTHETA] = o[1];
        Offset[aPHI] = o[2];
        for (v=aPSI; v<=aPHI; v++)  // convert offet mm/mrad to m/rad
            Offset[v] *= 0.001;
        if (oflg != TRUE) {
            Offset[aX]=0.;
            Offset[aY]=0.;
            Offset[aZ]=0.;
        }
        oflg = TRUE;
    }

    if (!rflg || !mflg || eflg) {
        msg("dataset (-r) and parameter file (-m) are required\n");
        do_help();  // doesn't return
    }

    CovName = Params.ParmName;
    p = get_parm("CovName");
    if (p->set) {
        CovName = (char *)p->ptr;
    }

    XformName = "";
    p = get_parm("XformName");
    if (p->set) {
        XformName = (char *)p->ptr;
    }

/*
    // check for model override
    i = 0;
    p = get_parm("doMultiSphere");
    if (p->set) {
        Params.Model = MSPHERE;
        i++;
    }
    p = get_parm("doNolte");
    if (p->set) {
        Params.Model = NOLTE;
        i++;
    }
    if (i == 2) {
        Cleanup("can't specify both MultiSphere and Nolte");
    }
*/

    // set the random number seed
    time(&secs);
    srandom((unsigned int)secs);

    // announce
    if (vflg) {
        printf("S A M   C O R E G I S T R A T I O N\t%s\n", PRG_REV);
        printf("opening data file\n");
        fflush(stdout);
    }

    if (vflg) {
        printf("opening data file");
        fflush(stdout);
    }

    // get data with sensor structures
    name = copy_string(DSName);
    GetDsInfo(DSName, &Header, &Channel, &Epoch, NULL, TRUE);
    GetDsInfo(name, &Header, &NewChannel, &Epoch, NULL, TRUE);
    GetFilePath(DSDIR, SAMpath, sizeof(SAMpath), &Params, "SAM", TRUE);
    free(name);

    if (Params.Model == MSPHERE) {
        if (vflg) {
            printf("opening multisphere model\n");
            fflush(stdout);
        }
        GetFilePath(DSDIR, fpath, sizeof(fpath), &Params, "default.hdm", TRUE);
        SetSphereOrigins(fpath, &Header, Channel, TRUE);    // read cortical hull model
        SetSphereOrigins(fpath, &Header, NewChannel, TRUE); // read cortical hull model
    }

    if (vflg) {
        printf(" - done\n");
    }

    // set dataset constants
    S = Header.NumSquid;
    M = Header.NumPri;

    p = get_parm("CovBand");
    if (!p->set) {
        fatalerr("CovBand req'd");
    }
    p = get_parm("MRIDirectory");
    if (!p->set) {
        fatalerr("MRIDirectory req'd");
    }
    p = get_parm("AtlasName");
    if (!p->set) {
        fatalerr("AtlasName req'd");
    }

    ORDER = Params.Order;
    //Tstep = Params.InitTrans;
    //Rstep = Params.InitRotat;

    // summarize coregistration parameters
    if (vflg) {
    if (Params.Model == MSPHERE) {
        printf("\tMultiSphere Model\n");
    } else if (Params.Model == NOLTE) {
            printf("\tNolte Order=%d\n", Params.Order);
        }
    }
    printf("\tNumber of Vertices=%d\n", Params.MaxVertex);
    //printf("\tInitial Translation=%5.3f mm\n", 1000.*Tstep);
    //printf("\tInitial Rotation=%5.3f mrad\n", 1000.*Rstep);
    //printf("\tEnding Translation=%5.3f mm\n", 1000.*Params.EndTrans);
    //printf("\tEnding Rotation=%5.3f mrad\n", 1000.*Params.EndRotat);
    printf("\tInitial Grid Extent=%5.3f mm\n", 1000.*Trans_max[0]);
    printf("\tInitial Grid Step=%5.3f mm\n", 1000.*Trans_step[0]);
    printf("\tInitial Rot Max=%5.3f mrad\n", 1000.*Rot_max[0]);
    printf("\tInitial Rot Step=%5.3f mrad\n", 1000.*Rot_step[0]);
    printf("\tFinal Grid Extent=%5.3f mm\n",1000.*Trans_max[4]);
    printf("\tFinal Grid Step=%5.3f mm\n", 1000.*Trans_step[4]);
    printf("\tFinal Rot Max=%5.3f mrad\n", 1000.*Rot_max[4]);
    printf("\tFinal Rot Step=%5.3f mrad\n", 1000.*Rot_step[4]);
    printf("\tIgnoring vertices below %7.3f cm\n", 100. * Params.MinZ);
    printf("\tMaximum angular error=%6.3f rad\n", Params.MaxAngle);
    printf("\tMaximum dot product=%5.3f\n", Params.MaxDot);
    printf("\tMinimum SNR=%5.3f x pass maximum SNR\n", Params.MinSNR);
    fflush(stdout);

    // set prefix to NumPrefix characters (default is 8-character dataset hashcode)
    memset(Prefix, 0, sizeof(Prefix));
    memcpy(Prefix, Header.SetName, Params.NumPrefix);


    if (Params.Model == NOLTE) {
        name = "hull.shape";
        p = get_parm("HullName");
        if (p->set) {
            name = (char *)p->ptr;
        }
        GetMRIPath(HullPath, sizeof(HullPath), &Params, name, TRUE);
        if (vflg) {
            printf("reading cortical hull model");
            fflush(stdout);
        }
        GetHull(HullPath, &Hull);           // read hull model
    }

    // get atlas
    if (vflg) {
        printf(" - done\n");
        printf("reading atlas file");
        fflush(stdout);
    }
    p = get_parm("AtlasName");
    name = (char *)p->ptr;
    GetMRIPath(AtlasPath, sizeof(AtlasPath), &Params, name, TRUE);

    // check for GIFTI atlas

    i = strlen(AtlasPath);
    if (i > 4 && strcmp(&AtlasPath[i - 4], ".gii") == 0) {
        gflg = TRUE;
    }

    // Read original atlas.

    if (gflg) {
        atlas = GetGiiAtlas(AtlasPath);
        Atlas = atlas->Voxel;
        Natlas = atlas->nrh + atlas->nlh;
        for (v=X_; v<=Z_; v++) {
            Vo[v] = atlas->cent[v];
        }
    } else {
        GetAtlas(AtlasPath, &Atlas, &Natlas);
    }
    if (vflg) {
        printf(" - read %d vertices", Natlas);
        fflush(stdout);
    }

    if (vflg) {
        printf("\nBounding box:\n");
        printf("\tx=%7.3f to %7.3f (cm)\n", 100. * vmin[X_], 100. * vmax[X_]);
        printf("\ty=%7.3f to %7.3f (cm)\n", 100. * vmin[Y_], 100. * vmax[Y_]);
        printf("\tz=%7.3f to %7.3f (cm)\n", 100. * vmin[Z_], 100. * vmax[Z_]);
        printf("Centroid:\n");
        printf("\t%7.3f, %7.3f, %7.3f (cm)\n", 100. * Vo[X_], 100. * Vo[Y_], 100. * Vo[Z_]);
        fflush(stdout);
    }

    // read covariance file & mean sensor noise
    if (vflg) {
        printf("reading covariance file & mean sensor noise\n");
        fflush(stdout);
    }
    snprintf(fpath, sizeof(fpath), "%s,%-d-%-dHz", CovName, (int)Params.CovHP, (int)Params.CovLP);
    GetFilePath("%d/SAM/%s/Global.cov", CovPath, sizeof(CovPath), &Params, fpath, 0);
    C = gsl_matrix_alloc(M, M);
    Cinv = gsl_matrix_alloc(M, M);
    GetCov(CovPath, &CovHeader, &ChanIndex, C);
    if (ndim > 0) {
        pinvN(C, Cinv, ndim);
    } else {
        pinv(C, Cinv);
    }

    name = copy_string(fpath);
    GetFilePath("%d/SAM/%s/Global_Noise", fpath, sizeof(fpath), &Params, name, 0);
    free(name);
    if ((fp = fopen(fpath, "r")) == NULL)
        Cleanup("can't open 'Global_Noise'");
    if (fscanf(fp, "%lf", &Noise) != 1)
        Cleanup("can't read 'Global_Noise'");
    fclose(fp);

    // open log file, & write header
    snprintf(fpath, sizeof(fpath), "%s,%sTransform.txt", Header.SetName,XformName);
    if ((fp = fopen(fpath, "w")) == NULL)
        Cleanup("can't open transform summary for write");
    fprintf(fp, "Dataset '%s'\n", Header.SetName);
    fprintf(fp, "Number of SQUID Sensors: %d (%d Primary Sensors)\n", S, M);
    fprintf(fp, "Cortical Mesh Coregistration Parameters:\n");
    if (Params.Model == MSPHERE) {
        fprintf(fp, "\tMultiSphere Model\n");
    } else if (Params.Model == NOLTE) {
        fprintf(fp, "\tNolte Order=%d\n", Params.Order);
    }
    fprintf(fp, "\tUsing %d random samples of %d Vertices\n", Params.MaxVertex, Natlas);
    //fprintf(fp, "\tInitial Translation=%5.3f mm\n", 1000.*Tstep);
    //fprintf(fp, "\tInitial Rotation=%5.3f mrad\n", 1000.*Rstep);
    //fprintf(fp, "\tEnding Translation=%5.3f mm\n", 1000.*Params.EndTrans);
    //fprintf(fp, "\tEnding Rotation=%5.3f mrad\n", 1000.*Params.EndRotat);
    fprintf(fp, "\tInitial Grid Extent=%5.3f mm\n", 1000.*Trans_max[0]);
    fprintf(fp, "\tInitial Grid Step=%5.3f mm\n", 1000.*Trans_step[0]);
    fprintf(fp, "\tInitial Rot Max=%5.3f mm\n", 1000.*Rot_max[0]);
    fprintf(fp, "\tInitial Rot Step=%5.3f mm\n", 1000.*Rot_step[0]);
    fprintf(fp, "\tFinal Grid Extent=%5.3f mm\n",1000.*Trans_max[4]);
    fprintf(fp, "\tFinal Grid Step=%5.3f mm\n", 1000.*Trans_step[4]);
    fprintf(fp, "\tFinal Rot Max=%5.3f mm\n", 1000*Rot_max[4]);
    fprintf(fp, "\tFinal Rot Step=%5.3f mm\n", 1000*Rot_step[4]);
    //end the Allison parameters
    fprintf(fp, "\tIgnoring vertices below %7.3f cm\n", 100. * Params.MinZ);
    fprintf(fp, "\tMaximum angular error=%6.3f rad\n", Params.MaxAngle);
    fprintf(fp, "\tMaximum dot product=%5.3f\n", Params.MaxDot);
    fprintf(fp, "\tMinimum SNR=%5.3f x pass maximum SNR\n", Params.MinSNR);
    fflush(fp);

    // allocate Mesh & Result structure
    if(Params.MaxVertex > 0) {          // select MaxVertex random mesh elements from N vertices
        Mesh = new_arrayE(MESHINFO, Params.MaxVertex, "Mesh[]");
        Atlas2RandomMesh(Atlas, Natlas, Mesh, &Params);
    } else {                            // select all vertices
        Mesh = new_arrayE(MESHINFO, Natlas, "Mesh[]");
        Atlas2Mesh(Atlas, Natlas, Mesh, &Params);
    }
    free(Atlas);
    if (vflg) {
        printf("selected %d Mesh vertices out of %d read \n", Params.MaxVertex, Natlas);
        fflush(stdout);
    }

    Best = new_arrayE(PASS_RESULTS, PASS_LIMIT, "Best[]");

    // initialize transform parameters
    for (v=X_; v<=Z_; v++) {
        To[v] = 0.;
        PriorTo[v] = 0.;
    }
    for (v=PSI_; v<=PHI_; v++) {
        Ro[v] = 0.;
        PriorRo[v] = 0.;
    }

    // generate scattergram using initial sensor positions
    if (vflg) {
        printf(" - done\n");
        printf("computing initial mesh scatter plot\n");
        fflush(stdout);
    }
    for (v=X_; v<=Z_; v++) {
        Where.Translate[v] = 0.;
        Rotation[v] = 0.;
    }

    if (oflg) { // modify 'Channel' positions if offsets are introduced
        if (vflg)
            printf("\nUsing offset %f %f %f %f %f %f\n",Offset[X_],Offset[Y_],Offset[Z_],Offset[aPSI],Offset[aTHETA],Offset[aPHI]); 
        Where.Translate[X_] = Offset[X_];
        Where.Translate[Y_] = Offset[Y_];
        Where.Translate[Z_] = Offset[Z_];
        Rotation[PSI_] = Offset[aPSI];
        Rotation[THETA_] = Offset[aTHETA];
        Rotation[PHI_] = Offset[aPHI];
        EtoR(Rotation, Where.Rotate);
        MoveMEG(Channel, Channel, Header.SqIndex, S, &Where);
        
    }

    EtoR(Rotation, Where.Rotate);
    MoveMEG(Channel, NewChannel, Header.SqIndex, S, &Where);

    if (Params.Model == MSPHERE) {
        MeshScatterMS(&Header, NewChannel, C, Cinv, Noise, Mesh, &Params, &Nused, &slope, &x2, &correlation, &rms);
    } else if (Params.Model == NOLTE) {
        MeshScatter(&Header, NewChannel, C, Cinv, Noise, &Hull, Mesh, &Params, &Nused, &slope, &x2, &correlation, &rms);
    }
    printf("x2: %f\t corr: %f Nused: %d MinSpan Used: %6.3f\n",x2,correlation,Nused,Params.MinSpan);
    // save initial scatter plot
    if (pflg) {
        sprintf(fpath, "%s_%s_00.txt", Prefix,XformName);
        if ((sfp = fopen(fpath, "w")) == NULL)
            Cleanup("can't open scatter plot file for write");
        i=0;
        for (n=0; n<Params.MaxVertex; n++)      // save angular error (radians) vs z^2
            if (Mesh[n].z2 > 0.) {
                fprintf(sfp, "%5.3f\t%6.3f\n", Mesh[n].angle, Mesh[n].z2);
	        i++;
        }
        fclose(sfp);
    }
    printf("Params.MaxVertex: %d  Nused: %d Vertices Written: %d\n",Params.MaxVertex, Nused,i);
    // step through transforms
    if (vflg) {
        printf(" - done\n");
        printf("evaluating transformations");
        fflush(stdout);
    }

    // initializations
    zprior = rms;           // initial rms S/N
    rprior = correlation;   // initial correlation
    sprior = slope;         // initial slope
    stp = 0;
    nstop = 0;

    for(passcounter=0; passcounter<1; passcounter++) {
        // log pass info
        printf("\nPass: %d\n", passcounter+1);
        fprintf(fp, "\nPass: %d\n", passcounter+1);
        fprintf(fp, "Max Grid Extent: %5.3f mm, %5.3f mrad\n", 1000*Trans_max[passcounter], 1000*Rot_max[passcounter]);
        fprintf(fp, "Step Sizes: %5.3f mm, %5.3f mrad\n", 1000*Trans_step[passcounter], 1000*Rot_step[passcounter]);
        fprintf(fp, "MinSpan Used: %6.3f \n",Params.MinSpan);
        fprintf(fp, "Current S/N: %6.3f rms\n", zprior);
        fprintf(fp, "Current Correlation: %9.6f\n", rprior);
        fprintf(fp, "Current Slope: %9.6f\n", sprior);
        fprintf(fp, "Initial Vertices Selected: %d\n", Nused);
        fprintf(fp, "Current Translation (x, y, z): %7.3f, %7.3f, %7.3f mm\n",
            1000.*To[X_], 1000.*To[Y_], 1000.*To[Z_]);
        fprintf(fp, "Current Rotation (psi, theta, phi): %7.3f, %7.3f, %7.3f mrad\n",
            1000.*Ro[PSI_], 1000.*Ro[THETA_], 1000.*Ro[PHI_]);
        if (passcounter > 0)
            fprintf(fp, "Distances: %9.6f mm, %9.6f mrad", 1000.*Tdisp, 1000.*Rdisp);
        switch (nstop) {
            case 0: fprintf(fp, "\n"); break;
            case 1: fprintf(fp, "\t*\n"); break;
            case 2: fprintf(fp, "\t**\n"); break;
            case 3: fprintf(fp, "\t***\n"); break;
            default: break;
        }

        // output pass translation header
        fprintf(fp, "\n                         Translation Evaluations\n");
        fprintf(fp, "      X       Y       Z     Psi   Theta     Phi     rms    chi-sq  Nused   correlation \n");
        fflush(fp);

        // (1) evaluate translation, only
#ifdef DOTRANS
        rbest = -1.;
        for (Translation[Z_]=(To[Z_]-Trans_max[passcounter]); Translation[Z_]<=(To[Z_]+Trans_max[passcounter]+SMIDGE); Translation[Z_]+=Trans_step[passcounter]) {
            for (Translation[Y_]=(To[Y_]-Trans_max[passcounter]); Translation[Y_]<=(To[Y_]+Trans_max[passcounter]+SMIDGE); Translation[Y_]+=Trans_step[passcounter]) {
                for (Translation[X_]=(To[X_]-Trans_max[passcounter]); Translation[X_]<=(To[X_]+Trans_max[passcounter]+SMIDGE); Translation[X_]+=Trans_step[passcounter]) {

                    printf("transforming sensor array\n");
                    // transform primary sensor array & reference sensors
                    for (v=X_; v<=Z_; v++)
                        Where.Translate[v] = Translation[v];
                    for (v=PSI_; v<=PHI_; v++)
                        Rotation[v] = Ro[v];
                    EtoR(Rotation, Where.Rotate);
                    MoveMEG(Channel, NewChannel, Header.SqIndex, S, &Where);
                    if (Params.Model == MSPHERE) {
                        MeshScatterMS(&Header, NewChannel, C, Cinv, Noise, Mesh, &Params, &Nused, &slope, &x2, &correlation, &rms);
                    } else if (Params.Model == NOLTE) {
                        MeshScatter(&Header, NewChannel, C, Cinv, Noise, &Hull, Mesh, &Params, &Nused, &slope, &x2, &correlation, &rms);
                    }

                    product = correlation * rms;
                    // find maximum correlation
                    if (correlation > rbest) {
                        zbest = rms;
                        rbest = correlation;
                        xbest = x2;
                        pbest = product;
                        for (v=X_; v<=Z_; v++)
                            Tmax[v] = Translation[v];
                    }
                    printf("correlation: %f\n",correlation);
                    //total_error=0;
                    //for (n=0; n<Params.MaxVertex; n++) {     //Calculate the total angular area
                    //    total_error = total_error + Mesh[n].angle;
                    //}
                    //printf("total error: %f\n",total_error);
                    //printf("Param.MaxVertex: %d\n",Params.MaxVertex);
                    //printf("mean_error: %f\n", mean_error);
                    //mean_error = total_error/Params.MaxVertex;

                    // log the current test transformation
                    fprintf(fp, "%7.3f\t%7.3f\t%7.3f\t%7.3f\t%7.3f\t%7.3f\t%7.3f\t%8.6f\t%4d\t%9.6f\t%9.6f\t%7.3f\n",
                      1000.*Translation[X_], 1000.*Translation[Y_], 1000.*Translation[Z_], 1000.*Ro[PSI_], 1000.*Ro[THETA_],
              1000.*Ro[PHI_], rms, x2, Nused, correlation, slope,Params.MinSpan);
                    fflush(fp);
                }       // end x-step
            }           // end y-step
        }               // end z-step

        // (2) pivot on best translation & save for averaging
        for (v=X_; v<=Z_; v++) {
            To[v] = Tmax[v];            // needed for rotatiions!
            Best[stp].To[v] = Tmax[v];
        }

        if (pflg) {
            sprintf(fpath, "%s_%s_01trans.txt", Prefix,XformName);
            if ((sfp = fopen(fpath, "w")) == NULL)
                Cleanup("can't open scatter plot file for write");
            for (n=0; n<Params.MaxVertex; n++)      // save angular error (radians) vs z^2
                if (Mesh[n].z2 > 0.)
                    fprintf(sfp, "%5.3f\t%6.3f\n", Mesh[n].angle, Mesh[n].z2);
            fclose(sfp);
        }
#endif
#ifdef DOROT
        // output rotation header
        fprintf(fp, "\n                         Rotation Evaluations\n");
        fprintf(fp, "      X       Y       Z     Psi   Theta     Phi     rms    x2  n  correlation slope\n");
        fflush(fp);

        // (3) evaluate rotation, only
        for(Rotation[PHI_]=(Ro[PHI_]-Rot_max[passcounter]); Rotation[PHI_]<=(Ro[PHI_]+Rot_max[passcounter]+SMIDGE); Rotation[PHI_]+=Rot_step[passcounter]) {
            for(Rotation[THETA_]=(Ro[THETA_]-Rot_max[passcounter]); Rotation[THETA_]<=(Ro[THETA_]+Rot_max[passcounter]+SMIDGE); Rotation[THETA_]+=Rot_step[passcounter]) {
                for(Rotation[PSI_]=(Ro[PSI_]-Rot_max[passcounter]); Rotation[PSI_]<=(Ro[PSI_]+Rot_max[passcounter]+SMIDGE); Rotation[PSI_]+=Rot_step[passcounter]) {

                    // transform primary sensor array & reference sensors
                    for (v=X_; v<=Z_; v++)
                        Where.Translate[v] = To[v];
                    EtoR(Rotation, Where.Rotate);
                    MoveMEG(Channel, NewChannel, Header.SqIndex, S, &Where);
                    if (Params.Model == MSPHERE) {
                        MeshScatterMS(&Header, NewChannel, C, Cinv, Noise, Mesh, &Params, &Nused, &slope, &x2, &correlation, &rms);
                    } else if (Params.Model == NOLTE) {
                        MeshScatter(&Header, NewChannel, C, Cinv, Noise, &Hull, Mesh, &Params, &Nused, &slope, &x2, &correlation, &rms);
                    }

                    product = correlation * rms;
                    // find maximum correlation
                    if (correlation > rbest) {
                        zbest = rms;
                        rbest = correlation;
                        xbest = x2;
                        sbest = slope;
                        pbest = product;
                        for (v=PSI_; v<=PHI_; v++)
                            Rmax[v] = Rotation[v];
                    }

                    //total_error=0;
                    //for (n=0; n<Params.MaxVertex; n++) {     //Calculate the total angular area
                    //    if (Mesh[n].z2 > 0.)
                    //        total_error = total_error + Mesh[n].angle;
                    //}
                    //mean_error = total_error/Params.MaxVertex;

                    // log the current test transformation
                    fprintf(fp, "%7.3f\t%7.3f\t%7.3f\t%7.3f\t%7.3f\t%7.3f\t%7.3f\t%8.6f\t%4d\t%9.6f\t%9.6f\t%7.3f\n",
                      1000.*To[X_], 1000.*To[Y_], 1000.*To[Z_], 1000.*Rotation[PSI_], 1000.*Rotation[THETA_],
              1000.*Rotation[PHI_], rms, x2, Nused, correlation, slope, Params.MinSpan);
                    fflush(fp);
                }       // end psi-step
            }           // end theta-step
        }               // end phi-step
#endif

        // (4) pivot on best rotation & save for averaging
        for (v=PSI_; v<=PHI_; v++) {
            Ro[v] = Rmax[v];
            Best[stp].Ro[v] = Rmax[v];
        }

        // (5) average all current translations
        //for (v=X_; v<=Z_; v++) {
        //    for (n=0, tmp=0.; n<=stp; n++)
        //        tmp += Best[n].To[v];
        //    To[v] = tmp / (double)(stp + 1);
        //}

        // (6) average all current rotations
        //for (v=PSI_; v<=PHI_; v++) {
        //    for (n=0, tmp=0.; n<=stp; n++)
        //        tmp += Best[n].Ro[v];
        //    Ro[v] = tmp / (double)(stp + 1);
        //}

        // (7) compute translation & rotation displacement from previous values
        for (v=X_, r2=0.; v<=Z_; v++) {
            tmp = To[v] - PriorTo[v];
            r2 += tmp * tmp;
        }
        Tdisp = sqrt(r2);
        for (v=PSI_, r2=0.; v<=PHI_; v++) {
            tmp = Ro[v] - PriorRo[v];
            r2 += tmp * tmp;
        }
        Rdisp = sqrt(r2);

        // (8) save prior values
        for (v=X_; v<=Z_; v++)
            PriorTo[v] = To[v];
        for (v=PSI_; v<=PHI_; v++)
            PriorRo[v] = Ro[v];
        zprior = zbest;
        rprior = rbest;

        // (9) test if both translation & rotation displacements are smaller than the stop conditions
        //if (Tdisp < (Tstep / 4.) && Rdisp < (Rstep / 4.))
        //    nstop++;        // count the number of times the displacements are smaller than criteria
        //if (nstop < 3) {
        //    stp++;
        //} else {
        //    Tstep /= 2.;    // reduce translation step size by factor of 2
        //    Rstep /= 2.;    // reduce rotation step size by factor of 2
        //    stp = 0;        // reset the average step counter
        //    nstop = 0;      // reset the stop count
        //}
        //pass++;             // bump pass count

        // (10) if req'd, save pass delta angle vs. z^2 scattergram for future analysis
        if (pflg) {
            sprintf(fpath, "%s_%s_01rot.txt", Prefix,XformName);
            if ((sfp = fopen(fpath, "w")) == NULL)
                Cleanup("can't open scatter plot file for write");
            for (n=0; n<Params.MaxVertex; n++)      // save angular error (radians) vs z^2
                if (Mesh[n].z2 > 0.)
                    fprintf(sfp, "%5.3f\t%6.3f\n", Mesh[n].angle, Mesh[n].z2);
            fclose(sfp);
        }

        // dot
        if (vflg) {
            printf(".");
            fflush(stdout);
        }

    } // ACN end pass for loop

    // report zbest, & translation

    fprintf(fp, "\nBest point Corr: %7.3f RMS: %7.3f x2: %7.3f slope: %9.6f product: %9.6f N: %d,at %7.3f, %7.3f, %7.3f mm, & %7.3f, %7.3f, %7.3f milliradians\n",
       rbest, zbest, xbest, sbest, pbest, Nused, 1000.*To[X_], 1000.*To[Y_], 1000.*To[Z_], 1000.*Ro[PSI_], 1000.*Ro[THETA_], 1000.*Ro[PHI_]);
    fclose(fp);

    // save transform parameters (unscaled) in SAM subdirectory
    sprintf(fname, "%s%s", XformName, Params.AtlasName);
    GetFilePath("%d/SAM/%s.xfm", fpath, sizeof(fpath), &Params, fname, 0);
    fp = fileopen(fpath, "w");
    fprintf(fp, "%e\t%e\t%e\t%e\t%e\t%e", To[X_], To[Y_], To[Z_], Ro[PSI_], Ro[THETA_], Ro[PHI_]);
    fclose(fp);

    // finit
    log_params(SAMpath);

    if (vflg) {
        printf(" - done\n");
        printf("'%s' done\n", Progname);
    }
    exit(0);
}
